"""
Test cases for SOHO EITL1Map subclass
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.mapbase import SpatialPair
from sunpy.map.sources.soho import EITL1Map
from .helpers import _test_private_date_setters


@pytest.fixture
def eit_l1_map():
    """Creates an EITL1Map"""
    return get_dummy_map_from_header(get_test_filepath("EIT_header/SOHO_EIT_171_20070601T120013_L1.header"))


def test_fitstoEIT(eit_l1_map):
    """Tests the creation of EITMap using FITS."""
    assert isinstance(eit_l1_map, EITL1Map)


def test_eitmap_coordinate_system(eit_l1_map):
    assert eit_l1_map.coordinate_system == SpatialPair(axis1='HPLN-TAN', axis2='HPLT-TAN')


def test_reference_date(eit_l1_map):
    assert eit_l1_map.reference_date.isot == "2007-06-01T11:59:05.180"


def test_date(eit_l1_map):
    assert eit_l1_map.date.isot == "2007-06-01T11:58:58.884"


def test_private_date_setters(eit_l1_map):
    _test_private_date_setters(eit_l1_map)


def test_is_datasource_for(eit_l1_map):
    """Test the is_datasource_for method of EITMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert eit_l1_map.is_datasource_for(eit_l1_map.data, eit_l1_map.meta)


def test_observatory(eit_l1_map):
    """Tests the observatory property of the EITMap object."""
    assert eit_l1_map.observatory == "Solar and Heliospheric Observatory (SOHO)"

def test_instrument(eit_l1_map):
    """Tests the instrument property of the EITMap object."""
    assert eit_l1_map.instrument == "Extreme-ultraviolet Imaging Telescope (EIT)"

def test_measurement(eit_l1_map):
    """Tests the measurement property of the EITMap object."""
    assert eit_l1_map.measurement == 171*u.angstrom


def test_rsun(eit_l1_map):
    """Tests the measurement property of the EITMap object."""
    assert u.allclose(eit_l1_map.rsun_obs, 953.98737017*u.arcsec)


def test_norm_clip(eit_l1_map):
    # Tests that the default normalizer has clipping disabled
    assert not eit_l1_map.plot_settings['norm'].clip


def test_wcs(eit_l1_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    eit_l1_map.pixel_to_world(0*u.pix, 0*u.pix)


def test_l1_eit_l1_map_colormaps():
    eit_l1_map = get_dummy_map_from_header(get_test_filepath("EIT_header/SOHO_EIT_171_20070601T120013_L1.header"))
    assert eit_l1_map.plot_settings["cmap"] == "sohoeit171"
    eit_l1_map = get_dummy_map_from_header(get_test_filepath("EIT_header/SOHO_EIT_195_20070601T121346_L1.header"))
    assert eit_l1_map.plot_settings["cmap"] == "sohoeit195"
    eit_l1_map = get_dummy_map_from_header(get_test_filepath("EIT_header/SOHO_EIT_284_20070601T120607_L1.header"))
    assert eit_l1_map.plot_settings["cmap"] == "sohoeit284"
    eit_l1_map = get_dummy_map_from_header(get_test_filepath("EIT_header/SOHO_EIT_304_20070601T121937_L1.header"))
    assert eit_l1_map.plot_settings["cmap"] == "sohoeit304"



"""Hinode XRT and SOT Map subclass definitions"""
import astropy.units as u
from astropy.visualization import ImageNormalize, LogStretch

from sunpy.map.mapbase import GenericMap, SpatialPair
from sunpy.map.sources.source_type import source_stretch

__all__ = ['XRTMap', 'SOTMap']


def _lower_list(alist):
    return [item.lower() for item in alist]


class XRTMap(GenericMap):
    """
    Hinode XRT map definition.

    The X-Ray Telescope (XRT) is a high resolution grazing incidence telescope,
    which is a successor to Yohkoh. It provides 2-arcsecond resolution images
    of the highest temperature solar coronal material,
    from 1,000,000 to 10,000,000 Kelvin.

    Hinode was launched on 22 September 2006 into a sun-synchronous orbit.

    Notes
    -----
    XRT files do not normally specify the heliographic longitude of the spacecraft,
    so sunpy silently assumes that the spacecraft is at zero Stonyhurst heliographic
    longitude (i.e., the same longitude as Earth) for L1 files. This assumption is
    safe for nearly all analyses due to Hinode's orbital altitude of only ~600 km.
    This assumption is not made if ``HGLN_OBS`` and ``HGLT_OBS`` values have been
    explicitly added to the metadata.

    References
    ----------
    * `Hinode Mission Page <https://solarb.msfc.nasa.gov/index.html>`__
    * `XRT Instrument Page <https://xrt.cfa.harvard.edu/>`__
    * `Fits header reference <https://hinode.nao.ac.jp/uploads/2016/04/22/SB_MW_Key13.pdf>`__
    * `Hinode User Guide <https://hinode.nao.ac.jp/en/for-researchers/analysis-guide/>`__
    * `XRT Analysis Guide <https://xrt.cfa.harvard.edu/science/tutorials.php>`__
    * `Coronal Temperature Diagnostic Capability of the Hinode/X-Ray Telescope Based on Self-Consistent Calibration <https://arxiv.org/abs/1011.2867>`__
    """
    filter_wheel1_measurements = ["Al_med", "Al_poly", "Be_med",
                                  "Be_thin", "C_poly", "Open"]
    filter_wheel2_measurements = ["Open", "Al_mesh", "Al_thick",
                                  "Be_thick", "Gband", "Ti_poly"]

    def __init__(self, data, header, **kwargs):
        fw1 = header.get('EC_FW1_', '')
        if fw1.lower() not in _lower_list(self.filter_wheel1_measurements):
            raise ValueError(f'Unexpected filter wheel 1 {fw1} in header.')
        fw2 = header.get('EC_FW2_', '')
        if fw2.lower() not in _lower_list(self.filter_wheel2_measurements):
            raise ValueError(f'Unexpected filter wheel 2 {fw2} in header.')
        super().__init__(data, header, **kwargs)
        self.plot_settings['cmap'] = 'hinodexrt'
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, LogStretch()), clip=False)

    @property
    def coordinate_system(self):
        """
        Override the default implementation to handle SOTMap-specific logic for CTYPE values.
        """
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1.lower() in ("solar-x", "solar_x"):
            ctype1 = 'HPLN-TAN'
        if ctype2.lower() in ("solar-y", "solar_y"):
            ctype2 = 'HPLT-TAN'
        return SpatialPair(ctype1, ctype2)

    @property
    def _timesys(self):
        if self.meta.get('timesys', '').upper() == 'UTC (TBR)':
            return 'UTC'
        return super()._timesys

    @property
    def _supported_observer_coordinates(self):
        # Assume observer is at zero Stonyhurst heliographic longitude if not otherwise specified
        # https://community.openastronomy.org/t/sunpymetadatawarnings-when-using-hinode-xrt-data/393/7 for more information
        return (super()._supported_observer_coordinates
                + [(('solar_b0', 'dsun_obs'), {'lon': 0*u.deg,
                                               'lat': self.meta.get('solar_b0'),
                                               'radius': self.meta.get('dsun_obs'),
                                               'unit': (u.deg, u.deg, u.m),
                                               'frame': "heliographic_stonyhurst"})])

    @property
    def detector(self):
        return "XRT"

    @property
    def observatory(self):
        return "Hinode"

    @property
    def measurement(self):
        fw1 = self.meta.get('EC_FW1_').replace("_", " ")
        fw2 = self.meta.get('EC_FW2_').replace("_", " ")
        return f"{fw1}-{fw2}"

    @property
    def processing_level(self):
        lvl = self.meta.get('DATA_LEV', None)
        if lvl is None:
            return
        return int(lvl)

    @property
    def unit(self):
        # XRT data values are in DN and are converted into DN/s if the data has been normalized.
        # A tag starting with "XRT_RENORMALIZE" is added to the HISTORY tag in that case.
        # See Table 1.1 and Section 2.11 of the XRT Analysis Guide.
        unit = super().unit
        if not unit:
            unit = u.DN
            history = self.meta.get('HISTORY', '')
            if "xrt_renormalize" in history.lower():
                unit = u.DN / u.second
        return unit

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an XRT image"""
        return header.get('instrume') == 'XRT'


class SOTMap(GenericMap):
    """Hinode SOT Image Map definition.

    The Hinode Solar Optical Telescope (SOT) consists of a 50 cm
    diffraction-limited Gregorian telescope. It is optimized for
    accurate measurement of the vector magnetic
    field in the photosphere and dynamics of both the photosphere and
    chromosphere associated with the magnetic fields.

    Hinode was launched on 22 September 2006 into a sun-synchronous orbit.

    References
    ----------
    * `Hinode Mission Page <https://solarb.msfc.nasa.gov/index.html>`__
    * `Hinode SOT Instrument Page <https://sot.lmsal.com>`__
    * `Hinode SOT Instrument Paper <https://arxiv.org/abs/0711.1715>`__
    * `Data Analsis Guide <https://sot.lmsal.com/doc/rep/sot254/fid366/SOT00042_33_SOT_Analysis_Guide_SAG.pdf>`__
    """
    # TODO: get a link for the SOT FITS headers
    # Add in some information about the the possible instrument, observation
    # type, observable ion and wavelength

    Instruments = ['SOT/WB', 'SOT/NB', 'SOT/SP', 'SOT/CT']

    Waves = ['6302A', 'BFI no move', 'CN bandhead 3883',
             'Ca II H line', 'G band 4305', 'NFI no move',
             'TF Fe I 6302', 'TF Mg I 5172', 'TF Na I 5896',
             'blue cont 4504', 'green cont 5550', 'red cont 6684']

    Observation_Type = ['FG (simple)', 'FG focus scan',
                        'FG shuttered I and V', 'FG shutterless I and V',
                        'FG shutterless I and V with 0.2s intervals',
                        'FG shutterless Stokes', 'SP IQUV 4D array']

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self._nickname = self.detector

        # TODO (add other options, Now all treated as intensity. This follows
        # Hinode SDC archive) StokesQUV -> grey, Velocity -> EIS, Width -> EIS,
        # Mag Field Azi -> IDL 5 (STD gamma II)
        # 'WB' -> red
        # 'NB'(0 = red); (>0 = gray), # nb has 1 stokes I, the rest quv
        # 'SP' (<=1 = red); (>1 = gray) #sp has 2 stokes I, the rest quv
        color = {'SOT/WB': 'intensity',
                 'SOT/NB': 'intensity',  # For the 1st dimension
                 'SOT/SP': 'intensity',  # For the 1st 2 dimensions
                 }

        self.plot_settings['cmap'] = 'hinodesot' + color[self.instrument]

    @property
    def coordinate_system(self):
        """
        Override the default implementation to handle SOTMap-specific logic for CTYPE values.
        """
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1.lower() in ("solar-x", "solar_x"):
            ctype1 = 'HPLN-TAN'
        if ctype2.lower() in ("solar-y", "solar_y"):
            ctype2 = 'HPLT-TAN'
        return SpatialPair(ctype1, ctype2)

    @property
    def detector(self):
        return "SOT"

    @property
    def observatory(self):
        return "Hinode"

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an SOT image."""
        return header.get('instrume') in cls.Instruments



"""
This module provides a generalized dictionary class that deals with header
parsing, normalization, and maintaining coherence between keys and keycomments.
"""
from collections import OrderedDict, namedtuple
from collections.abc import Mapping

__all__ = ['MetaDict']

ModifiedItem = namedtuple('ModifiedItem', ['original', 'current'])
ModifiedItem.__repr__ = lambda t: f"(original={t.original}, current={t.current})"


class MetaDict(OrderedDict):
    """
    A class to hold metadata associated with a `sunpy.map.Map` derivative.

    This class handles everything in lower case. This allows case
    insensitive indexing.

    If the key 'keycomments' exists, its value must be a dictionary mapping
    keys in the :class:`MetaDict` to their comments. The casing of keys in the
    keycomments dictionary is not significant. If a key is removed from the
    :class:`MetaDict`, it will also be removed from the keycomments dictionary.
    Additionally, any extraneous keycomments will be removed when the
    :class:`MetaDict` is instantiated.

    Parameters
    ----------
    save_original : bool, optional
        If `True` (the default), a copy of the original metadata will be saved to the
        `original_meta` property. Note that this keyword argument is
        required as an implementation detail to avoid recursion when saving
        the original contents.
    """

    def __init__(self, *args, save_original=True):
        # Store all keys as lower-case to allow for case-insensitive indexing
        # OrderedDict can be instantiated from a list of lists or a tuple of tuples
        tags = dict()
        if args:
            args = list(args)
            adict = args[0]

            if isinstance(adict, MetaDict):
                self._original_meta = adict._original_meta

            if isinstance(adict, list) or isinstance(adict, tuple):
                items = adict
            elif isinstance(adict, Mapping):
                # Cast to a dict here, since a Mapping doesn't have to define
                # a .items() method (but has enough methods to be converted to
                # a dict)
                items = dict(adict).items()
            else:
                raise TypeError(f"Can not create a MetaDict from this input of type {type(adict)}")

            self._check_str_keys(items)
            tags = OrderedDict((k.lower(), v) for k, v in items)
            args[0] = tags

        super().__init__(*args)
        # Use `copy=True` to avoid mutating the caller's keycomments
        # dictionary (if they provided one).
        self._prune_keycomments(copy=True)

        if save_original and not hasattr(self, '_original_meta'):
            self._original_meta = MetaDict(*args, save_original=False)

    def __str__(self):
        return '\n'.join([f"('{key}': '{item}')" for key, item in self.items()])

    def __repr__(self):
        return f"{self.__class__.__name__}([{self}])"

    # Deliberately a property to prevent external modification
    @property
    def original_meta(self):
        """
        A copy of the dictionary from when it was initialised.
        """
        return self._original_meta

    @property
    def added_items(self):
        """
        Items added since initialisation.
        """
        return {k: self[k] for k in set(self) - set(self.original_meta)}

    @property
    def removed_items(self):
        """
        Items removed since initialisation.
        """
        return {k: self.original_meta[k] for k in set(self.original_meta) - set(self)}

    @property
    def modified_items(self):
        """
        Items modified since initialisation.

        This is a mapping from ``key`` to ``[original_value, current_value]``.
        """
        keys = set(self).intersection(set(self.original_meta))
        return {k: ModifiedItem(self.original_meta[k], self[k]) for k in keys if
                self[k] != self.original_meta[k]}

    def copy(self):
        """
        Shallow copies this instance.
        """
        copied = super().copy()
        # By default the above line will overwrite original_meta, so manually re-instate it
        copied._original_meta = self.original_meta
        return copied

    @staticmethod
    def _check_str_keys(items):
        bad_keys = []
        for k, v in items:
            if not isinstance(k, str):
                bad_keys.append(str(k))
        if len(bad_keys):
            raise ValueError('All MetaDict keys must be strings, '
                             'found the following non-compliant keys: ' +
                             ', '.join(bad_keys))

    def _prune_keycomments(self, copy=False):
        """
        Remove keycomments for keys that are not contained in the MetaDict.

        Parameters
        ----------
        copy : `bool`, optional
            Make a copy of the current keycomments dict before removing keys.
        """
        if 'keycomments' not in self:
            return

        keycomments = self['keycomments']

        if not isinstance(keycomments, dict):
            raise TypeError(
                "'keycomments' key must have a value of type `dict`. Found "
                f"the following type: {type(keycomments)!r}")

        if copy:
            keycomments = keycomments.copy()

        for key in list(keycomments.keys()):
            if key not in self:
                del keycomments[key]

        self['keycomments'] = keycomments

    def __contains__(self, key):
        """
        Override ``__contains__``.
        """
        return OrderedDict.__contains__(self, key.lower())

    def __getitem__(self, key):
        """
        Override ``[]`` indexing.
        """
        return OrderedDict.__getitem__(self, key.lower())

    def __setitem__(self, key, value):
        """
        Override ``[]`` indexing.
        """
        return OrderedDict.__setitem__(self, key.lower(), value)

    def popitem(self, last):
        key, value = super().popitem(last)
        self._prune_keycomments()
        return key, value

    def __delitem__(self, key):
        """
        Override ``del dict[key]`` key deletion.
        """
        OrderedDict.__delitem__(self, key.lower())
        self._prune_keycomments()

    def item_hash(self):
        """
        Create a hash based on the stored items.

        This relies on all the items themselves being hashable. For this reason
        the 'keycomments' item, which is a dict, is excluded from the hash.

        If creating the hash fails, returns `None`.
        """
        self_copy = self.copy()
        self_copy.pop('keycomments', None)
        try:
            return hash(frozenset(self_copy.items()))
        except Exception:
            return

    def get(self, key, default=None):
        """
        Override ``.get()`` indexing.
        """
        return OrderedDict.get(self, key.lower(), default)

    def has_key(self, key):
        """
        Override ``.has_key()`` to perform case-insensitively.
        """
        return key.lower() in self

    def pop(self, key, default=None):
        """
        Override ``.pop()`` to perform case-insensitively.
        """
        has_key = key in self
        result = OrderedDict.pop(self, key.lower(), default)
        if has_key:
            self._prune_keycomments()
        return result

    def update(self, d2):
        """
        Override ``.update()`` to perform case-insensitively.
        """
        return OrderedDict.update(self, OrderedDict((k.lower(), v) for k, v in d2.items()))

    def setdefault(self, key, default=None):
        """
        Override ``.setdefault()`` to perform case-insensitively.
        """
        return OrderedDict.setdefault(self, key.lower(), default)



"""
===========================
Loading an HMI synoptic map
===========================

In this example we load a synoptic map produced by the HMI team. This
data is an interesting demonstration of sunpy's Map class as it is not in the
more common Helioprojective coordinate system, but in heliographic Carrington
coordinates and a cylindrical equal area (CEA) projection.
"""
import matplotlib.pyplot as plt

from astropy.utils.data import download_file

import sunpy.map

###############################################################################
# Download the file and read it into a Map.

filename = download_file(
    'http://jsoc.stanford.edu/data/hmi/synoptic/hmi.Synoptic_Mr.2191.fits', cache=True)
syn_map = sunpy.map.Map(filename)

###############################################################################
# Plot the results.

fig = plt.figure(figsize=(12, 5))
ax = plt.subplot(projection=syn_map)
im = syn_map.plot(axes=ax)

ax.coords[0].set_axislabel("Carrington Longitude [deg]")
ax.coords[1].set_axislabel("Latitude [deg]")

ax.coords.grid(color='black', alpha=0.6, linestyle='dotted', linewidth=0.5)

cb = plt.colorbar(im, fraction=0.019, pad=0.1)
cb.set_label(f"Radial magnetic field [{syn_map.unit}]")

# In order to make the x-axis ticks show, the bottom y-limit has to be adjusted slightly
ax.set_ylim(bottom=0)
ax.set_title(f"{syn_map.meta['content']},\n"
             f"Carrington rotation {syn_map.meta['CAR_ROT']}")

plt.show()



"""Yohkoh SXT Map subclass definitions"""
import astropy.units as u
from astropy.coordinates import ITRS, SphericalRepresentation
from astropy.visualization import PowerStretch
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.map import GenericMap
from sunpy.map.sources.source_type import source_stretch

__all__ = ['SXTMap']


class SXTMap(GenericMap):
    """Yohkoh SXT Image Map

    The Yohkoh Soft X-ray Telescope (SXT) observed the full solar disk
    (42 x 42 arcminutes) in the 0.25 - 4.0 keV range.
    It consisted of a glancing incidence mirror and a CCD sensor and
    used thin metallic filters to acquire images in restricted
    portions of its energy range.
    SXT could resolve features down to 2.5 arcseconds.
    Information about the temperature and density of the plasma
    emitting the observed x-rays was obtained by comparing images acquired with
    the different filters.
    Images could be obtained every 2 to 8 seconds.
    Smaller images with a single filter could be obtained as frequently as
    once every 0.5 seconds.
    Yohkoh was launched on 30 August 1991 and ceased operations on
    14 December 2001.

    Notes
    -----
    Observer location:  We use the ITRS coordinates provided in
    the FITS header (``LAT``, ``LONG``, ``RADIUS``)
    for the spacecraft location even when coordinates in the heliographic
    Stonyhurst frame are provided (``HGS_LON``, ``HGS_LAT``, ``DSUN_OBS``).
    The two locations differ substantially, and the ITRS coordinates
    are more accurate than the HGS information.

    Using this information for the observer coordinate was implemented
    in ``sunpy`` 4.0.6 to work with the FITS files that were re-processed in 2016.

    References
    ----------
    * `Yohkoh Mission Page <https://solar.physics.montana.edu/sxt/>`__
    * `Fits header reference <https://proba2.sidc.be/data/SWAP/level0>`__
    * `Yohkoh Analysis Guide <http://ylstone.physics.montana.edu/ylegacy/yag.html>`__
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)

        self.plot_settings['cmap'] = 'yohkohsxt' + self.measurement[0:2].lower()
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, PowerStretch(0.5)), clip=False)

    @property
    def _supported_observer_coordinates(self):
        return [(('long', 'lat', 'radius'), {'lon': self.meta.get('long'),
                                             'lat': self.meta.get('lat'),
                                             'distance': self.meta.get('radius'),
                                             'unit': (u.deg, u.deg, u.km),
                                             'representation_type': SphericalRepresentation,
                                             'frame': ITRS, })
                ] + super()._supported_observer_coordinates

    @property
    def observatory(self):
        return "Yohkoh"

    @property
    def detector(self):
        return "SXT"

    @property
    def measurement(self):
        s = self.meta.get('wavelnth', '')
        if s == 'Al.1':
            s = 'Al01'
        elif s.lower() == 'open':
            s = 'white-light'
        return s

    @property
    def wavelength(self):
        """
        Returns `None`, as SXT is a broadband imager.
        """

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an SXT image"""
        return header.get('instrume') == 'SXT'



"""
========================================================================
Creating a TimeSeries from GOES-XRS near real time data with flare times
========================================================================

This example will demonstrate how to make use of GOES XRS Near Real Time data.
This includes the goes XRS timeseries data as well as the flare times.
The real-time datasets can all be found at https://services.swpc.noaa.gov/json/goes/primary/
"""

import matplotlib.pyplot as plt
import pandas as pd

from astropy import units as u

from sunpy import timeseries as ts
from sunpy.time import parse_time

###############################################################################
# We will start by getting reading the GOES-XRS JSON file using :func:`pandas.read_json`.
# This allows us to download the file and load it straight into a `pandas.DataFrame`.
# This file updates every minute and contains the last 7 days worth of data.
# There is also a file with only the last 3 days worth of data.

goes_data = pd.read_json(
    "https://services.swpc.noaa.gov/json/goes/primary/xrays-7-day.json"
)

###############################################################################
# The recorded flux values alternate between the two XRS energy channels:
# "0.05-0.4nm" and "0.1-0.8nm". We make a `pivot table <https://en.wikipedia.org/wiki/Pivot_table>`__
# that naturally rearranges the data into two flux columns. We then rename the
# columns.

goes_data = goes_data.pivot(
    index='time_tag',
    columns='energy',
    values='observed_flux'
)
goes_data.rename(columns={'0.05-0.4nm': 'xrsa', '0.1-0.8nm': 'xrsb'}, inplace=True)

###############################################################################
# `sunpy.timeseries.TimeSeries` requires a datetime index, which we can get by
# parsing the time strings.

goes_data.index = parse_time(goes_data.index).datetime

###############################################################################
# `sunpy.timeseries.TimeSeries` requires that there are units for data variables.
# To do this, we will create a dictionary that will map the names of the two columns,
# "xrsa" and "xrsb" (the channel names for GOES XRS), to their corresponding
# physical flux units, ``u.W/u.m**2``.

units = dict([("xrsa", u.W / u.m ** 2), ("xrsb", u.W / u.m ** 2)])

###############################################################################
# We need to create a metadata dictionary for the data.
# Typically, `sunpy.timeseries.TimeSeries` reads the metadata directly from the file.
# However, here we need to define our own metadata and we will keep it fairly simple.

meta = dict(
    {"instrument": "GOES X-ray sensor", "measurements": "primary", "type": "quicklook"}
)

###############################################################################
# Now we will create a `sunpy.timeseries.TimeSeries` by passing in the data,
# the metadata and the units.

goes_ts = ts.TimeSeries(goes_data, meta, units, source="xrs")

###############################################################################
# NOAA also provides the past 7 days of flare event identifications which we
# can also parse.
# Note that it is possible that no flares occurred in the last 7 days.

flare_events = pd.read_json(
    "https://services.swpc.noaa.gov/json/goes/primary/xray-flares-7-day.json"
)

###############################################################################
# Next we sort the list and retain the five largest flares.

if len(flare_events) > 0:
    largest_flares = flare_events.sort_values('max_xrlong', ascending=False)[:5]
else:
    largest_flares = []

###############################################################################
# Finally, we plot the timeseries and overlay the largest flares that occurred.

fig, ax = plt.subplots()
goes_ts.plot(axes=ax)
if len(largest_flares) > 0:
    for _, this_flare in largest_flares.iterrows():
        peak_time = parse_time(this_flare["max_time"])
        ax.axvline(peak_time.datetime, label=f'{peak_time.datetime} {this_flare["max_class"]}')
ax.legend()
ax.set_title(f"Last 7 days of GOES XRS data with the largest {len(largest_flares)} flares labeled")
plt.show()



from sunpy.data.data_manager.cache import Cache
from sunpy.data.data_manager.downloader import ParfiveDownloader
from sunpy.data.data_manager.manager import DataManager
from sunpy.data.data_manager.storage import SqliteStorage

__all__ = ['Cache', 'DataManager', 'ParfiveDownloader', 'SqliteStorage']



"""
Test cases for PUNCH subclass.
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.punch import PUNCHMap


@pytest.fixture
def punch_map():
    return get_dummy_map_from_header(get_test_filepath("punch.header"))


def test_punch_map(punch_map):
    """Tests the creation of PUNCHMap"""
    assert isinstance(punch_map, PUNCHMap)


def test_reference_date(punch_map):
    """ Tests the reference_date"""
    assert punch_map.reference_date.isot == "2025-03-12T03:26:00.000"


def test_date(punch_map):
    assert punch_map.date.isot == "2025-03-12T03:26:00.000"


def test_unit(punch_map):
    assert punch_map.unit == u.Unit("2.009e+07 W / (sr m2)")


def test_is_datasource_for(punch_map):
    """Tests the is_datasource_for method of PUNCHMap."""
    assert punch_map.is_datasource_for(punch_map.data, punch_map.meta)


def test_observatory(punch_map):
    """Tests the observatory property of the PUNCHMap object."""
    assert punch_map.observatory == "PUNCH"


def test_instrument(punch_map):
    """Tests the instrument property of the PUNCHMap object."""
    assert punch_map.instrument == "WFI+NFI Mosaic"


def test_wcs(punch_map):
    """Test WCS is valid and can transform from pixels to world coordinates"""
    punch_map.pixel_to_world(0*u.pix, 0*u.pix)



from sunpy.net.dataretriever import GenericClient

__all__ = ['GBMClient']


class GBMClient(GenericClient):
    """
    Provides access to data from the Gamma-Ray Burst Monitor (GBM) instrument
    on board the Fermi satellite.

    Although GBMs primary objective is to detect gamma-ray bursts,
    it provides high quality high energy solar flare observations.

    The instrument consists of 12 Sodium Iodide (NaI) scintillation
    detectors, which are sensitive to an energy range of 4keV to 1MeV.
    At any one time, 6 of the NaI detectors are Sunward facing.
    The detectors are numbered 'n1' to 'n11'. This client supports the user
    to choose which detector to use through the `a.Detector <sunpy.net.attrs.Detector>` attribute.
    The default detector is 'n5'.

    The GBM data comes in daily version files in two formats:

        * CSPEC - counts accumulated every  4.096 seconds in 128 energy channels for each detector.
        * CTIME - counts accumulated every 0.256 seconds in 8 energy channels

    Both of which can be accessed through the attrs `a.Resolution <sunpy.net.attrs.Resolution>`.
    The default data type is CSPEC unless the user defines.

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a
    >>> res = Fido.search(a.Time('2015-06-21 00:00', '2015-06-23 23:59'),
    ...                   a.Instrument.gbm, a.Detector.n3,
    ...                   a.Resolution.ctime) # doctest: +REMOTE_DATA
    >>> res # doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    3 Results from the GBMClient:
    Source: https://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/daily
    <BLANKLINE>
           Start Time               End Time        ... Resolution Detector
    ----------------------- ----------------------- ... ---------- --------
    2015-06-21 00:00:00.000 2015-06-21 23:59:59.999 ...      ctime       n3
    2015-06-22 00:00:00.000 2015-06-22 23:59:59.999 ...      ctime       n3
    2015-06-23 00:00:00.000 2015-06-23 23:59:59.999 ...      ctime       n3
    <BLANKLINE>
    <BLANKLINE>

    """

    pattern = ('https://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/daily/{{year:4d}}/{{month:2d}}/{{day:2d}}/current/'
                'glg_{{Resolution:5}}_{{Detector:2}}_{{year:2d}}{{month:2d}}{{day:2d}}_v00.pha')

    @property
    def info_url(self):
        return 'https://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/daily'

    @classmethod
    def register_values(cls):
        from sunpy.net import attrs
        adict = {attrs.Instrument: [('GBM', 'Gamma-Ray Burst Monitor on board the Fermi satellite.')],
                 attrs.Physobs: [('flux', 'a measure of the amount of radiation received by an object from a given source.')],
                 attrs.Source: [('FERMI', 'The Fermi Gamma-ray Space Telescope.')],
                 attrs.Provider: [('NASA', 'The National Aeronautics and Space Administration.')],
                 attrs.Resolution: [
            ("cspec", "CSPEC 128 channel spectra every 4.096 seconds."),
            ("ctime", "CTIME provides 8 channel spectra every 0.256 seconds.")],
            attrs.Detector: [(f"n{x}", f"GBM Detector short name for the detector NAI_{x:02}") for x in range(12)]}
        return adict



import pathlib
import functools
from contextlib import contextmanager
from urllib.parse import urlparse

from sunpy.util.exceptions import warn_user
from sunpy.util.util import hash_file

__all__ = ['DataManager']


class DataManager:
    """
    This class provides a remote data manager for managing remote files.

    Parameters
    ----------
    cache: `sunpy.data.data_manager.cache.Cache`
        Cache object to be used by `~sunpy.data.data_manager.manager.DataManager`.
    """

    def __init__(self, cache):
        self._cache = cache

        self._file_cache = {}

        self._namespace = None
        self._skip_hash_check = False
        self._skip_file = {}
        self._require_files = {}

    def require(self, name, urls, sha_hash, defer_download=False):
        """
        Decorator for informing the data manager about the requirement of
        a file by a function. Optionally defer downloading the file until it's
        requested in `~sunpy.data.data_manager.manager.DataManager.get`.

        Parameters
        ----------
        name : `str`
            The name to reference the file with.
        urls : `list` or `str`
            A list of urls to download the file from.
        sha_hash : `str`
            SHA-256 hash of file.
        defer_download : `bool`, optional
            If `True`, the file download is deferred until it is requested via `~sunpy.data.data_manager.manager.DataManager.get`.
        """
        if isinstance(urls, str):
            urls = [urls]

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                self._namespace = self._get_module(func)
                if defer_download:
                    self._require_files[name] = {
                        'urls': urls,
                        'sha_hash': sha_hash,
                    }
                else:
                    self._download_and_cache_file(name, urls, sha_hash)

                result = func(*args, **kwargs)
                self._namespace = None
                return result
            return wrapper

        return decorator

    def _download_and_cache_file(self, name, urls, sha_hash):
        """
        Internal method to handle the downloading and caching logic.
        """
        replace = self._skip_file.get(name)
        if replace:
            uri_parse = urlparse(replace['uri'])
            if uri_parse.scheme in ("", "file"):
                file_path = uri_parse.netloc + uri_parse.path
                file_hash = hash_file(file_path)
            else:
                file_path, file_hash, _ = self._cache._download_and_hash([replace['uri']], self._namespace)
            if replace['hash'] and file_hash != replace['hash']:
                raise ValueError("Hash provided to override_file does not match hash of the file.")
        elif self._skip_hash_check:
            file_path = self._cache.download(urls, self._namespace, redownload=True)
        else:
            details = self._cache.get_by_hash(sha_hash)
            if not details:
                if self._cache_has_file(urls):
                    raise ValueError(f"{urls} has already been downloaded, but no file matching the hash {sha_hash} can be found.")
                file_path = self._cache.download(urls, self._namespace)
                file_hash = hash_file(file_path)
                if file_hash != sha_hash:
                    raise RuntimeError(f"Hash of local file ({file_hash}) does not match expected hash ({sha_hash}). File may have changed on the remote server.")
            else:
                if not pathlib.Path(details["file_path"]).is_file():
                    warn_user("Requested file appears to missing and will be redownloaded.")
                    self._cache._download_and_hash(urls, self._namespace)

                if hash_file(details['file_path']) != details['file_hash']:
                    warn_user("Hashes do not match, the file will be redownloaded (could be tampered/corrupted)")
                    file_path = self._cache.download(urls, self._namespace, redownload=True)
                    if hash_file(file_path) != details['file_hash']:
                        raise RuntimeError("Redownloaded file also has the incorrect hash. The remote file on the server might have changed.")
                else:
                    file_path = details['file_path']

        if name not in self._file_cache:
            self._file_cache[name] = {}
        self._file_cache[name][self._namespace] = file_path

    def get(self, name):
        """
        Get the file by name, and download it if deferred.

        Parameters
        ----------
        name : `str`
            Name of the file given to the data manager, same as the one provided
            in `~sunpy.data.data_manager.manager.DataManager.require`.

        Returns
        -------
        `pathlib.Path`
            Path of the file.

        Raises
        ------
        `KeyError`
            If ``name`` is not in the cache.
        """
        if name in self._require_files:
            file_info = self._require_files.pop(name)
            self._download_and_cache_file(name, file_info['urls'], file_info['sha_hash'])

        return pathlib.Path(self._file_cache[name][self._namespace])

    @contextmanager
    def override_file(self, name, uri, sha_hash=None):
        """
        Replaces the file by the name with the file provided by the url/path.

        Parameters
        ----------
        name : `str`
            Name of the file provided in the `require` decorator.
        uri : `str`
            URI of the file which replaces original file. Scheme should be one
            of ``http``, ``https``, ``ftp`` or ``file``. If no scheme is given
            the uri will be interpreted as a local path. i.e.
            ``file:///tmp/test`` and ``/tmp/test`` are the same.
        sha_hash : `str`, optional
            SHA256 hash of the file to compared to after downloading.
        """
        try:
            self._skip_file[name] = {
                'uri': uri,
                'hash': sha_hash,
            }
            yield
        finally:
            _ = self._skip_file.pop(name, None)

    @contextmanager
    def skip_hash_check(self):
        """
        Disables hash checking temporarily.

        Examples
        --------
        >>> with remote_data_manager.skip_hash_check():  # doctest: +SKIP
        ...     myfunction()  # doctest: +SKIP
        """
        try:
            self._skip_hash_check = True
            yield
        finally:
            self._skip_hash_check = False

    def _cache_has_file(self, urls):
        for url in urls:
            if self._cache._get_by_url(url):
                return True
        return False

    def _get_module(self, func):
        """
        Returns the name of the module (appended with a dot) that a function belongs to.

        Parameters
        ----------
        func : function
            A function whose module is to be found.

        Returns
        -------
        A `str` that represents the module name appended with a dot.
        """
        module = func.__module__.lstrip('.').split('.')[0] + '.'
        if module == '__main__.':
            module = ''
        return module



"""
Datasource-specific classes

This is where datasource specific logic is implemented. Each mission should
have its own file with one or more classes defined. Typically, these classes
will be subclasses of the :mod`sunpy.map.Map` class.
"""

from sunpy.map.map_factory import Map
from .adapt import *
from .asos import *
from .gong import *
from .hinode import *
from .iris import *
from .mlso import *
from .proba2 import *
from .psp import *
from .punch import *
from .rhessi import *
from .sdo import *
from .soho import *
from .solo import *
from .source_type import *
from .stereo import *
from .suit import *
from .suvi import *
from .trace import *
from .yohkoh import *



"""
Test cases for KCor Map subclass.
"""
import pytest

import astropy.units as u
from astropy.tests.helper import assert_quantity_allclose

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.mlso import KCorMap
from .helpers import _test_private_date_setters


@pytest.fixture()
def kcor():
    """Creates an KCorMap from a FITS file."""
    return get_dummy_map_from_header(get_test_filepath("20181209_180305_kcor_l2.header"))


def test_kcormap_creation(kcor):
    """Tests the creation of KCorMap using FITS."""
    assert isinstance(kcor, KCorMap)


def test_is_datasource_for(kcor):
    """Test the is_datasource_for method of KCorMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert kcor.is_datasource_for(kcor.data, kcor.meta)


def test_measurement(kcor):
    """Tests the measurement property of the KCorMap object."""
    assert kcor.measurement == 735 * u.nm


def test_observatory(kcor):
    """Tests the observatory property of the KCorMap object."""
    assert kcor.observatory == "MLSO"


def test_norm_clip(kcor):
    # Tests that the default normalizer has clipping disabled
    assert not kcor.plot_settings['norm'].clip


def test_reference_date(kcor):
    assert kcor.reference_date.isot == "2018-12-09T18:03:05.000"


def test_date(kcor):
    assert kcor.date.isot == "2018-12-09T18:03:05.000"


def test_private_date_setters(kcor):
    _test_private_date_setters(kcor)


def test_wcs(kcor):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    kcor.pixel_to_world(0*u.pix, 0*u.pix)


def test_observer_coordinate(kcor):
    assert 'dsun_obs' not in kcor.meta

    # The test header triggers using the default observer coordinate
    observer = kcor.observer_coordinate.itrs.earth_location
    assert_quantity_allclose(observer.lon, kcor._earth_location.lon)
    assert_quantity_allclose(observer.lat, kcor._earth_location.lat)
    assert_quantity_allclose(observer.height, kcor._earth_location.height)

    # The test header will have a fully specified observer when DSUN_OBS is added
    kcor.meta['dsun_obs'] = (1*u.AU).to_value(u.m)
    kcor = KCorMap(kcor.data, kcor.meta)

    # The observer coordinate should now no longer be the default observer coordinate
    assert_quantity_allclose(kcor.observer_coordinate.radius, 1*u.AU)
    assert_quantity_allclose(kcor.observer_coordinate.lat, kcor.meta['crlt_obs']*u.deg)



import numpy as np
import pytest
from hypothesis import given
from parfive import Downloader

import astropy.units as u

from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.net._attrs import Time
from sunpy.net.dataretriever.client import QueryResponse
from sunpy.net.dataretriever.sources.aia_synopsis import AIASynopsisClient
from sunpy.net.fido_factory import UnifiedResponse
from sunpy.net.tests.strategies import time_attr
from sunpy.time.timerange import TimeRange


@pytest.fixture
def client():
    return AIASynopsisClient()


@pytest.mark.remote_data
@pytest.mark.parametrize(("timerange", "url_start", "url_end"), [
    (Time('2024/4/21', '2024/4/21 00:01'),
     'https://jsoc1.stanford.edu/data/aia/synoptic/2024/04/21/H0000/AIA20240421_0000_0094.fits',
     'https://jsoc1.stanford.edu/data/aia/synoptic/2024/04/21/H0000/AIA20240421_0000_4500.fits'
     ),
    (Time('2024/5/5', '2024/5/5 00:02'),
     'https://jsoc1.stanford.edu/data/aia/synoptic/2024/05/05/H0000/AIA20240505_0000_0094.fits',
     'https://jsoc1.stanford.edu/data/aia/synoptic/2024/05/05/H0000/AIA20240505_0002_1700.fits',
     ),
])
def test_get_url_for_time_range(client, timerange, url_start, url_end):
    qresponse = client.search(timerange)
    urls = [i['url'] for i in qresponse]
    assert isinstance(urls, list)
    assert urls[0] == url_start
    assert urls[-1] == url_end


@given(time_attr())
def test_can_handle_query(time):
    client = AIASynopsisClient()
    # This is to avoid the horrific representation
    ans = client._can_handle_query(time, a.Instrument.aia, a.Level("1.5s"))
    assert ans
    ans = client._can_handle_query(time, a.Instrument.aia)
    assert not ans
    ans = client._can_handle_query(time, a.Instrument.aia, a.Resolution(4.0))
    assert not ans
    ans = client._can_handle_query(time)
    assert not ans


@pytest.mark.remote_data
def test_query(client):
    time_range = TimeRange("2020-01-01", "2020-01-01 00:01")
    query_result = client.search(
        a.Time(time_range.start, time_range.end),
        a.Wavelength(171 * u.angstrom),
        a.Sample(12 * u.hour),
        a.Instrument("AIA"),
        a.Level("1.5s"),
    )
    assert query_result is not None

    qr1 = client.search(Time('2024/8/9', '2024/8/9 00:01'), a.Level("1.5s"))
    assert isinstance(qr1, QueryResponse)
    assert len(qr1) == 9
    assert qr1['Start Time'][0].isot == "2024-08-09T00:00:00.000"
    assert qr1['Start Time'][-1].isot == "2024-08-09T00:00:00.000"


@pytest.mark.remote_data
def test_wavelength_query(client):
    time_range = TimeRange("2020-01-01", "2020-01-01 00:12")
    query_result = client.search(
        a.Time(time_range.start, time_range.end),
        a.Instrument("AIA"),
        a.Level("1.5s"),
        a.Wavelength(1600 * u.angstrom),
    )
    assert len(query_result) == 7
    assert np.all(query_result["Wavelength"] == 1600)

    query_result_meters = client.search(
        a.Time(time_range.start, time_range.end),
        a.Instrument("AIA"),
        a.Level("1.5s"),
        a.Wavelength((1600*1e-10) * u.meter),
    )
    assert len(query_result_meters) == len(query_result)
    assert np.all(query_result_meters["Wavelength"] == query_result["Wavelength"])

    query_result_ghz = client.search(
        a.Time(time_range.start, time_range.end),
        a.Instrument("AIA"),
        a.Level("1.5s"),
        a.Wavelength(1873702.8625*u.GHz),
    )
    assert len(query_result_ghz) == len(query_result)
    assert np.all(query_result_ghz["Wavelength"] == query_result["Wavelength"])


@pytest.mark.remote_data
def test_sample_query(client):
    time_range = TimeRange("2020-01-01", "2020-01-02")
    query_result = client.search(
        a.Time(time_range.start, time_range.end),
        a.Wavelength(171 * u.angstrom),
        a.Instrument("AIA"),
        a.Level("1.5s"),
        a.Sample(12 * u.hour),
    )
    assert np.all(query_result["Wavelength"] == 171)
    assert len(query_result) == 3


@pytest.mark.remote_data
def test_get(client):
    qr1 = client.search(Time('2024/8/9', '2024/8/9 00:00:30'), a.Level("1.5s"))
    res = client.fetch(qr1, downloader=Downloader(max_conn=1, max_splits=1))
    assert len(res) == len(qr1)

    res2 = client.fetch(qr1[0], downloader=Downloader(max_conn=1, max_splits=1))
    assert len(res2) == 1


@pytest.mark.remote_data
def test_fido(tmpdir):
    time_range = TimeRange("2020-01-01", "2020-01-01 00:00:12")
    query_result = Fido.search(
        a.Time(time_range.start, time_range.end),
        a.Instrument.aia,
        a.Level("1.5s"),
    )
    assert len(query_result) == 1
    assert len(query_result[0]) == 9
    assert query_result is not None
    assert not np.all(query_result[0]["Wavelength"] == 1600)

    assert isinstance(query_result, UnifiedResponse)
    assert isinstance(query_result[0].client, AIASynopsisClient)

    response = Fido.fetch(query_result, path=tmpdir, downloader=Downloader(max_conn=1, max_splits=1))
    assert len(response) == query_result._numfile


def test_attr_reg():
    assert a.Level.onepointfive_s is not None
    assert a.Level.onepointfive_s.value == '1.5s'


def test_client_repr(client):
    output = str(client)
    assert output[:50] == 'sunpy.net.dataretriever.sources.aia_synopsis.AIASy'



import numpy as np
import pytest
from packaging.version import Version

import asdf
import astropy.units as u
from asdf.testing.helpers import roundtrip_object

import sunpy.map
from sunpy.data.test import get_test_filepath
from sunpy.tests.helpers import asdf_entry_points


def assert_roundtrip_map(old):
    new = roundtrip_object(old)
    np.testing.assert_allclose(old.data, new.data)

    # Test the meta by force!
    for ok, ov in old.meta.items():
        assert ok in new.meta
        assert new.meta[ok] == ov

    if old.mask is not None and new.mask is not None:
        np.testing.assert_allclose(old.mask, new.mask)

    assert old.unit == new.unit


def asdf_open_memory_mapping_kwarg(memmap: bool) -> dict:
    if Version(asdf.__version__) >= Version("3.1.0"):
        return {"memmap": memmap}
    else :
        return {"copy_arrays": not memmap}


@pytest.fixture
def aia171_test_map():
    aia_path = get_test_filepath("aia_171_level1.fits")
    return sunpy.map.Map(aia_path)


@asdf_entry_points
def test_genericmap_basic(aia171_test_map, tmpdir):
    assert_roundtrip_map(aia171_test_map)


@asdf_entry_points
def test_genericmap_mask(aia171_test_map, tmpdir):
    mask = np.zeros_like(aia171_test_map.data)
    mask[10, 10] = 1
    aia171_test_map.mask = mask
    aia171_test_map._unit = u.m
    assert_roundtrip_map(aia171_test_map)


@asdf_entry_points
def test_load_100_file_with_shift():
    fname = get_test_filepath("aiamap_shift_genericmap_1.0.0.asdf")
    with asdf.open(fname, **asdf_open_memory_mapping_kwarg(memmap=False)) as af:
        aiamap = af['object']
        assert isinstance(aiamap, sunpy.map.sources.AIAMap)
        assert "crval1" in aiamap.meta.modified_items
        crval1 = aiamap.meta.modified_items["crval1"]
        assert crval1.current - crval1.original == 10


@asdf_entry_points
def test_load_100_file_with_no_shift():
    fname = get_test_filepath("aiamap_genericmap_1.0.0.asdf")
    with asdf.open(fname, **asdf_open_memory_mapping_kwarg(memmap=False)) as af:
        aiamap = af['object']
        assert isinstance(aiamap, sunpy.map.sources.AIAMap)
        assert "crval1" not in aiamap.meta.modified_items
        assert "crval2" not in aiamap.meta.modified_items



"""
==============
Plotting a map
==============

How to create a plot of a map.
"""
# sphinx_gallery_thumbnail_number = 2
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE

###############################################################################
# We start with the sample data.

aiamap = sunpy.map.Map(AIA_171_IMAGE)

##############################################################################
# Let's plot the result. Setting the projection is necessary to ensure that
# pixels can be converted accurately to coordinates values.

fig = plt.figure()
ax = fig.add_subplot(projection=aiamap)
aiamap.plot(axes=ax)
aiamap.draw_limb(axes=ax)
aiamap.draw_grid(axes=ax)
plt.show()

##############################################################################
# The above image looks "dark" because the color scale is accounting for the
# small set of pixels that are extremely bright. We can use the keyword
# ``clip_interval`` to clip out pixels with extreme values. Here, we clip out
# the darkest 1% of pixels and the brightest 0.01% of pixels.

fig = plt.figure()
ax = fig.add_subplot(projection=aiamap)
aiamap.plot(axes=ax, clip_interval=(1, 99.99)*u.percent)
aiamap.draw_limb(axes=ax)
aiamap.draw_grid(axes=ax)
plt.show()



from packaging.requirements import Requirement

from sunpy.util.sysinfo import (
    find_dependencies,
    format_requirement_string,
    missing_dependencies_by_extra,
    resolve_requirement_versions,
    system_info,
)

EXTRA_DEPS = [
    "astropy",
    "beautifulsoup4",
    "cdflib",
    "drms",
    "h5netcdf",
    "h5py",
    "lxml",
    "matplotlib",
    "mpl-animators",
    "numpy",
    "packaging",
    "pandas",
    "parfive",
    "pyerfa",
    "python-dateutil",
    "reproject",
    "scipy",
    "zeep",
]

EXTRA_ALL_GROUPS = [
    "all",
    "asdf",
    "dev",
    "docs-gallery",
    "docs",
    "image",
    "jpeg2000",
    "map",
    "net",
    "opencv",
    "required",
    "spice",
    "tests",
    "timeseries",
    "visualization",
]

def test_find_dependencies():
    """
    This is ran in several test environments with varying dependencies installed.
    So it will be common to find not docs installed, so there will be "missing" dependencies.
    But this is not a problem.
    """
    _, installed = find_dependencies(package="sunpy", extras=["required", *EXTRA_ALL_GROUPS])
    for package in EXTRA_DEPS:
        assert package in installed

    # There should not be any self-referential dependency on sunpy
    assert "sunpy" not in installed


def test_missing_dependencies_by_extra():
    missing = missing_dependencies_by_extra()
    for group in EXTRA_ALL_GROUPS:
        assert group in missing


def test_resolve_requirement_versions():
    package1 = Requirement('test-package[ext1]>=1.1.1; extra == "group1"')
    package2 = Requirement('test-package[ext2]<=2.0.0; extra == "group2"')
    assert str(resolve_requirement_versions([package1, package2])) == str(Requirement(
        'test-package[ext1,ext2]<=2.0.0,>=1.1.1; extra == "group1" or extra == "group2"'))

    package3 = Requirement('test-package==1.1.0; extra == "group3"')
    package4 = Requirement('test-package==1.1.0; extra == "group4"')
    assert str(resolve_requirement_versions([package3, package4])) == str(
        Requirement('test-package==1.1.0; extra == "group3" or extra == "group4"'))

    package5 = Requirement('test-package; extra == "group5"')
    package6 = Requirement('test-package[ext3]@https://foo.com')
    assert str(resolve_requirement_versions([package5, package6])) == str(
        Requirement('test-package[ext3]@ https://foo.com ; extra == "group5"'))


def test_format_requirement_string():
    package1 = Requirement('test-package[ext1]>=1.1.1; extra == "group1"')
    assert format_requirement_string(package1) == 'Missing test-package[ext1]>=1.1.1; extra == "group1"'

    package2 = Requirement('test-package>=1.1.1; extra == "group1" or extra == "group2" or extra == "group3"')
    assert format_requirement_string(
        package2) == 'Missing test-package>=1.1.1; extra == "group1" or "group2" or "group3"'

    package3 = Requirement('test-package>=1.1.1')
    assert format_requirement_string(package3) == 'Missing test-package>=1.1.1'


def test_system_info(capsys):
    system_info()
    captured = capsys.readouterr()
    lines = captured.out.splitlines()
    assert "sunpy Installation Information" in lines

    # sunpy should not be listed as an optional dependency
    index = lines.index("Optional Dependencies")
    for line in lines[index + 2:]:
        assert not line.startswith("sunpy:")



"""
======================
Blending maps together
======================

This example shows how to blend two maps.

``matplotlib`` by itself provides only alpha-based transparency for
superimposing one image onto another, which may not be visually appealing.
For better results, one can manipulate the rendered image data arrays to perform a
standard image-compositing `blend mode <https://en.wikipedia.org/wiki/Blend_modes>`__.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
import sunpy.map
from sunpy.coordinates import SphericalScreen

###############################################################################
# Let's load two maps for blending. We reproject the second map to the
# coordinate frame of the first map for proper compositing, taking care to use
# the :class:`~sunpy.coordinates.SphericalScreen` context manager in order to
# preserve off-disk data.

a171 = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
a131 = sunpy.map.Map(sunpy.data.sample.AIA_131_IMAGE)
with SphericalScreen(a171.observer_coordinate):
    a131 = a131.reproject_to(a171.wcs)

###############################################################################
# Let's first plot the two maps individually to check the desired plot
# settings.

fig1 = plt.figure(figsize=(10, 4))
ax1 = fig1.add_subplot(121, projection=a171)
ax2 = fig1.add_subplot(122, projection=a131)

a171.plot(axes=ax1, clip_interval=(1, 99.999)*u.percent)
a131.plot(axes=ax2, clip_interval=(1, 99.9)*u.percent)

###############################################################################
# Now let's prepare for the computation of the blended image. We plot both
# maps using the desired plot settings in order to obtain the ``matplotlib``
# image instances for future use.

fig2 = plt.figure()
ax = fig2.add_subplot(projection=a171)

im171 = a171.plot(axes=ax, clip_interval=(1, 99.99)*u.percent)
im131 = a131.plot(axes=ax, clip_interval=(1, 99.9)*u.percent)

# sphinx_gallery_defer_figures

###############################################################################
# We use those ``matplotlib`` image instances to obtain RGB data arrays after
# normalization (including clipping) and the color maps have been applied. We
# need to provide the current ``matplotlib`` renderer and turn off resampling
# to screen pixels. To facilitate later arithmetic, we then scale the data
# arrays from being integers ranging 0-255 to being floats ranging 0.0-1.0.

renderer = fig2.canvas.get_renderer()
rgb171 = im171.make_image(renderer, unsampled=True)[0] / 255.
rgb131 = im131.make_image(renderer, unsampled=True)[0] / 255.

# sphinx_gallery_defer_figures

###############################################################################
# With these RGB arrays, we can now perform the calculation for the
# `screen blend mode <https://en.wikipedia.org/wiki/Blend_modes#Screen>`__.
# A different blend mode would simply require a different expression.

rgb_composite = 1 - (1 - rgb171) * (1 - rgb131)

# sphinx_gallery_defer_figures

###############################################################################
# Finally, we plot the composite image on the same axes.

ax.imshow(rgb_composite, origin='lower')
ax.set_title('Composite using the screen blend mode')
plt.show()

# sphinx_gallery_thumbnail_number = 2



from datetime import datetime

import pytest
from dateutil.relativedelta import relativedelta

from sunpy.net.scraper_utils import date_floor, extract_timestep, get_timerange_from_exdict
from sunpy.time import TimeRange, parse_time

DATETIME_PATTERN_EXAMPLES = [
    ('%b%y', relativedelta(months=1)),
    ('%m%y', relativedelta(months=1)),
    ('%H%d', relativedelta(hours=1)),
    ('%y%b', relativedelta(months=1)),
]

@pytest.mark.parametrize(('pattern', 'mintime'), DATETIME_PATTERN_EXAMPLES)
def test_extract_timestep(pattern, mintime):
    assert mintime == extract_timestep(pattern)

@pytest.mark.parametrize(('exdict', 'start', 'end'), [
    ({"year": 2000}, '2000-01-01 00:00:00', '2000-12-31 23:59:59.999000'),
    ({"year": 2016, "month": 2}, '2016-02-01 00:00:00', '2016-02-29 23:59:59.999000'),
    ({'year': 2019, 'month': 2, 'day': 28}, '2019-02-28 00:00:00', '2019-02-28 23:59:59.999000'),
    ({'year': 2019, 'month': 2, 'day': 28, 'hour': 23}, '2019-02-28 23:00:00', '2019-02-28 23:59:59.999000'),
    ({'year': 2019, 'month': 2, 'day': 28, 'hour': 23, 'minute': 59},
     '2019-02-28 23:59:00', '2019-02-28 23:59:59.999000'),
    ({'year': 2020, 'month': 7, 'day': 31, 'hour': 23, 'minute': 59, 'second': 59},
     '2020-07-31 23:59:59', '2020-07-31 23:59:59.999000'),
])
def test_get_timerange_from_exdict(exdict, start, end):
    tr = TimeRange(start, end)
    file_timerange = get_timerange_from_exdict(exdict)
    assert file_timerange == tr

@pytest.mark.parametrize(('testdate', 'pattern', 'floor_val'), [
    ((2004, 3, 6), '%y', datetime(2004, 1, 1, 0, 0)),
    ((2004, 3, 6), '%b%y', datetime(2004, 3, 1, 0, 0)),
    ((2023, 3, 12), '%y%M', datetime(2023, 3, 12, 0, 0)),
    ((2023, 2, 3), '%m%d', datetime(2023, 2, 3, 0, 0)),
    ((2023, 2, 23), '%d%H', datetime(2023, 2, 23, 0, 0)),
    ((2023, 12, 12), '%H%S', datetime(2023, 12, 12, 0, 0)),
])
def test_date_floor(testdate, pattern, floor_val):
    date = parse_time(testdate)
    timestep = extract_timestep(pattern)
    assert date_floor(date, timestep) == floor_val



from io import BytesIO

from asv_runner.benchmarks.mark import SkipNotImplemented, skip_benchmark
from matplotlib.figure import Figure

import astropy.units as u

import sunpy.data.sample
import sunpy.map
from sunpy.coordinates import propagate_with_solar_surface


class Creation:
    params = ['AIA_171_IMAGE', 'HMI_LOS_IMAGE']
    param_names = ['name']

    def setup(self, name):
        self.filename = getattr(sunpy.data.sample, name)

    def time_create_map(self, name):
        sunpy.map.Map(self.filename)

    # Skipped due to a bug in pympler.asizeof
    # https://github.com/pympler/pympler/issues/151
    @skip_benchmark
    def mem_create_map(self, name):
        return sunpy.map.Map(self.filename)

    def peakmem_create_map(self, name):
        sunpy.map.Map(self.filename)


class Resample:
    def setup_cache(self):
        aiamap = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
        return aiamap

    def time_resample(self, aiamap):
        aiamap.resample([100, 100] * u.pix)

    def peakmem_resample(self, aiamap):
        aiamap.resample([100, 100] * u.pix)


class Rotate:
    params = (['scipy', 'scikit-image', 'opencv'], range(0, 6))
    param_names = ['method', 'order']

    def setup_cache(self):
        aiamap = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
        return aiamap

    def setup(self, aiamap, method, order):
        if method == 'opencv' and order not in {0, 1, 3}:
            raise SkipNotImplemented

    def time_rotate(self, aiamap, method, order):
        aiamap.rotate(30*u.deg, method=method, order=order)

    def peakmem_rotate(self, aiamap, method, order):
        aiamap.rotate(30*u.deg, method=method, order=order)


class Reproject:
    params = ['interpolation', 'adaptive']
    param_names = ['algorithm']

    def setup_cache(self):
        maps = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE, sunpy.data.sample.HMI_LOS_IMAGE)
        return maps

    def time_reproject_to(self, maps, algorithm):
        maps[1].reproject_to(maps[0].wcs, algorithm=algorithm)

    def peakmem_reproject_to(self, maps, algorithm):
        maps[1].reproject_to(maps[0].wcs, algorithm=algorithm)

    def time_reproject_to_plus_diffrot(self, maps, algorithm):
        with propagate_with_solar_surface():
            maps[1].reproject_to(maps[0].wcs, algorithm=algorithm)

    def peakmem_reproject_to_plus_diffrot(self, maps, algorithm):
        with propagate_with_solar_surface():
            maps[1].reproject_to(maps[0].wcs, algorithm=algorithm)


class Autoalign:
    params = [False, True, 'pcolormesh']
    param_names = ['autoalign']

    def setup_cache(self):
        maps = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE, sunpy.data.sample.HMI_LOS_IMAGE)
        return maps

    def time_autoalign(self, maps, autoalign):
        fig = Figure()
        ax = fig.add_subplot(projection=maps[0])
        maps[1].plot(axes=ax, autoalign=autoalign)
        buf = BytesIO()
        fig.savefig(buf, format='png')

    def peakmem_autoalign(self, maps, autoalign):
        fig = Figure()
        ax = fig.add_subplot(projection=maps[0])
        maps[1].plot(axes=ax, autoalign=autoalign)
        buf = BytesIO()
        fig.savefig(buf, format='png')



import glymur
from scipy.ndimage import zoom


def resize_jp2(jp2file: str, scale: float) -> None:
    """
    Resize a JPEG2000 file by a given scale factor.

    It uses zoom from scipy.ndimage to resize the image data, so factor is the inverse.

    It will copy the XML metadata from the original file to the new one.
    It will not overwrite the original file.
    """
    original_jp2 = glymur.Jp2k(jp2file)
    rescaled_data = zoom(original_jp2[:], scale)
    new_file = glymur.Jp2k(f'{jp2file}_rescale.jp2', rescaled_data)
    for box in original_jp2.box:
        if isinstance(box, glymur.jp2box.XMLBox):
            new_file.append(box)



"""
Screen class definitions for making assumptions about off-disk emission
"""
import abc
import logging

import numpy as np

import astropy.units as u
from astropy.coordinates.representation import CartesianRepresentation, UnitSphericalRepresentation

from sunpy import log
from sunpy.coordinates import HeliographicStonyhurst, Helioprojective, _transformations
from sunpy.util.decorators import ACTIVE_CONTEXTS
from sunpy.util.exceptions import warn_user

__all__ = ['BaseScreen', 'SphericalScreen', 'PlanarScreen']


class BaseScreen(abc.ABC):

    def __init__(self, only_off_disk=False):
        self.only_off_disk = only_off_disk

    @property
    def _context_name(self):
        # Used for setting active context
        return f"{self.__module__}.{self.__class__.__qualname__}"

    @abc.abstractmethod
    @u.quantity_input
    def calculate_distance(self) -> u.cm:
        ...

    def __enter__(self):
        ACTIVE_CONTEXTS.append(self._context_name)
        self._old_assumed_screen = Helioprojective._assumed_screen  # nominally None
        Helioprojective._assumed_screen = self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if (removed := ACTIVE_CONTEXTS.pop()) != self._context_name:
            raise RuntimeError(f"Cannot remove {self._context_name} from tracking stack because {removed} is last active.")
        Helioprojective._assumed_screen = self._old_assumed_screen

    def _iterate_calculate_distance(self, coord, distance, screen_frame):
        """
        Numerically calculates the distance component to a screen to promote a
        coordinate from 2D to 3D.

        This method should be used only if the screen has been differentially rotated,
        since otherwise promoting a coordinate from 2D to 3D is straightforward.

        The approach used for numerical solving is to iterate to the solution starting
        with an initial guess. For each guess, we construct a 3D coordinate and perform
        a 3D->2D->3D transformation in the native frame of the screen. The amount that
        the coordinate shifts in the 3D->2D->3D transformation is related to how far off
        the guess is: when the shift is zero, the guess is correct. To form the next
        guess, we linearly interpolate/extrapolate between the latest guess and the
        closest guess over past iterations to estimate where the shift will be zero.

        The maximum number of iterations is hard-coded to 20, but convergence will
        typically take no more than 8 iterations. When working with an array coordinate,
        elements that have converged are skipped in subsequent iterations.
        """
        # Check if the logging level is at least DEBUG (for performance reasons)
        debug_output = log.getEffectiveLevel() <= logging.DEBUG

        log.debug("Differentially rotating the screen")
        use = np.ones(distance.shape, dtype=bool)
        delta = np.empty_like(distance)
        for niter in range(20):
            # If this is not the first iteration, update best distance and calculate new distance
            if niter > 0:
                if niter == 1:
                    # For the second iteration, use a simple guess
                    best_distance, distance = distance, distance + delta
                    best_delta = delta.copy()
                else:
                    last_distance = distance.copy()
                    # Starting with the third iteration, linearly interpolate between the last distance and the best distance
                    with np.errstate(invalid='ignore'):
                        distance[use] = last_distance[use] - delta[use] * np.nan_to_num((best_distance[use] - last_distance[use]) / (best_delta[use] - delta[use]))

                    # Update the best distance (not including the most recent calculation)
                    best_distance = np.where(np.abs(best_delta) < np.abs(delta), best_distance, last_distance)
                    best_delta = np.where(np.abs(best_delta) < np.abs(delta), best_delta, delta)

            # Calculate the corresponding delta of a 3D->2D->3D transformation in the native frame of the screen
            # If delta is zero in that frame, then the 3D point is exactly on the screen, so the guessed distance is correct
            other_3d = coord[use].realize_frame(coord[use].represent_as(UnitSphericalRepresentation) * distance[use])
            native_3d = other_3d.transform_to(screen_frame)
            native_2d = native_3d.realize_frame(native_3d.represent_as(UnitSphericalRepresentation))
            delta[use] = self.calculate_distance(native_2d) - native_3d.spherical.distance

            # Continue computations on only those elements that do not meet the tolerance
            # A tolerance of 1e-11 is larger than numerical-precision errors (~1e-12), and only 1.5 meters at 1 AU
            with np.errstate(invalid='ignore'):
                use = np.abs(delta / distance) >= 1e-11*u.one

            if debug_output:
                argmax = np.unravel_index(np.argmax(np.abs(delta)), delta.shape)
                log.debug("%d points remaining; largest delta is %s at distance %s", np.sum(use), delta[argmax], distance[argmax])

            # Return the solution if all elements have converged
            if np.sum(use) == 0:
                log.debug("Solved for the differentially rotated screen after %d iterations", niter + 1)
                return distance

            # Squash small deltas to avoid warnings from later computations
            delta = np.where(use, delta, 0)

        warn_user(f"Failed to solve for the differentially rotated screen after {niter + 1} iterations. "
                  "Using the best guess.")
        return distance


class SphericalScreen(BaseScreen):
    """
    Context manager to interpret 2D coordinates as being on the inside of a spherical screen.

    The radius of the screen is the distance between the specified ``center`` and Sun center.
    This ``center`` does not have to be the same as the observer location for the coordinate
    frame. If they are the same, then this context manager is equivalent to assuming that the
    helioprojective "zeta" component is zero.
    This replaces the default assumption where 2D coordinates are mapped onto the surface of the
    Sun.

    .. note:: This applies only to coordinates in a `~sunpy.coordinates.Helioprojective` frame.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The center of the spherical screen
    radius : `~astropy.units.Quantity`, optional
        The radius of the spherical screen. The default sets the radius to the distance from the
        screen center to the Sun.
    only_off_disk : `bool`, optional
        If `True`, apply this assumption only to off-disk coordinates, with on-disk coordinates
        still mapped onto the surface of the Sun. Defaults to `False`.

    See Also
    --------
    sunpy.coordinates.PlanarScreen

    Notes
    -----
    If this context manager is combined with the :func:`~sunpy.coordinates.propagate_with_solar_surface`
    context manager, significantly more computations are required to numerically solve for the
    distance to the differentially rotated screen.

    Examples
    --------
    .. minigallery:: sunpy.coordinates.SphericalScreen

    >>> import astropy.units as u
    >>> from sunpy.coordinates import Helioprojective, SphericalScreen
    >>> h = Helioprojective(range(7)*u.arcsec*319, [0]*7*u.arcsec,
    ...                     observer='earth', obstime='2020-04-08')
    >>> print(h.make_3d())
    <Helioprojective Coordinate (obstime=2020-04-08T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        [(   0., 0., 0.99660825), ( 319., 0., 0.99687244),
            ( 638., 0., 0.99778472), ( 957., 0., 1.00103285),
            (1276., 0.,        nan), (1595., 0.,        nan),
            (1914., 0.,        nan)]>

    >>> with SphericalScreen(h.observer):
    ...     print(h.make_3d())
    <Helioprojective Coordinate (obstime=2020-04-08T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        [(   0., 0., 1.00125872), ( 319., 0., 1.00125872),
            ( 638., 0., 1.00125872), ( 957., 0., 1.00125872),
            (1276., 0., 1.00125872), (1595., 0., 1.00125872),
            (1914., 0., 1.00125872)]>

    >>> with SphericalScreen(h.observer, only_off_disk=True):
    ...     print(h.make_3d())
    <Helioprojective Coordinate (obstime=2020-04-08T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        [(   0., 0., 0.99660825), ( 319., 0., 0.99687244),
            ( 638., 0., 0.99778472), ( 957., 0., 1.00103285),
            (1276., 0., 1.00125872), (1595., 0., 1.00125872),
            (1914., 0., 1.00125872)]>
    """
    screen_type = 'spherical'

    @u.quantity_input
    def __init__(self, center, *, radius: u.m=None, **kwargs):
        self._center = center
        if radius is not None:
            self._radius = radius
        else:
            hgs_frame = HeliographicStonyhurst(obstime=self._center.obstime)
            center_hgs = self._center.transform_to(hgs_frame)
            self._radius = center_hgs.radius
        super().__init__(**kwargs)

    def calculate_distance(self, frame):
        sphere_center = self._center.transform_to(frame).cartesian
        c = sphere_center.norm()**2 - self._radius**2
        rep = frame.represent_as(UnitSphericalRepresentation)
        b = -2 * sphere_center.dot(rep)
        # Ignore sqrt of NaNs
        with np.errstate(invalid='ignore'):
            distance = ((-1*b) + np.sqrt(b**2 - 4*c)) / 2  # use the "far" solution

        # Iterate the calculation if differential rotation is being applied
        if _transformations._autoapply_diffrot is not None and np.any(frame.obstime != self._center.obstime):
            screen_frame = Helioprojective(observer=self._center, obstime=self._center.obstime)
            distance = self._iterate_calculate_distance(frame, distance, screen_frame)

        return distance


class PlanarScreen(BaseScreen):
    """
    Context manager to interpret 2D coordinates as being on the inside of a planar screen.

    The plane goes through Sun center and is perpendicular to the vector between the
    specified vantage point and Sun center.
    This replaces the default assumption where 2D coordinates are mapped onto the surface of the
    Sun.

    .. note:: This applies only to coordinates in a `~sunpy.coordinates.Helioprojective` frame.

    Parameters
    ----------
    vantage_point : `~astropy.coordinates.SkyCoord`
        The vantage point that defines the orientation of the plane.
    distance_from_center : `~astropy.units.Quantity`, optional
        Distance from Sun center of the planar screen. Defaults to 0 such
        that the plane goes through Sun center
    only_off_disk : `bool`, optional
        If `True`, apply this assumption only to off-disk coordinates, with on-disk coordinates
        still mapped onto the surface of the Sun. Defaults to `False`.

    See Also
    --------
    sunpy.coordinates.SphericalScreen

    Notes
    -----
    If this context manager is combined with the :func:`~sunpy.coordinates.propagate_with_solar_surface`
    context manager, significantly more computations are required to numerically solve for the
    distance to the differentially rotated screen.

    Examples
    --------

    >>> import astropy.units as u
    >>> from sunpy.coordinates import Helioprojective, PlanarScreen
    >>> h = Helioprojective(range(7)*u.arcsec*319, [0]*7*u.arcsec,
    ...                     observer='earth', obstime='2020-04-08')
    >>> print(h.make_3d())
    <Helioprojective Coordinate (obstime=2020-04-08T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        [(   0., 0., 0.99660825), ( 319., 0., 0.99687244),
            ( 638., 0., 0.99778472), ( 957., 0., 1.00103285),
            (1276., 0.,        nan), (1595., 0.,        nan),
            (1914., 0.,        nan)]>

    >>> with PlanarScreen(h.observer):
    ...     print(h.make_3d())
    <Helioprojective Coordinate (obstime=2020-04-08T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        [(   0., 0., 1.00125872), ( 319., 0., 1.00125992),
            ( 638., 0., 1.00126351), ( 957., 0., 1.0012695 ),
            (1276., 0., 1.00127788), (1595., 0., 1.00128866),
            (1914., 0., 1.00130183)]>

    >>> with PlanarScreen(h.observer, distance_from_center=1*u.R_sun):
    ...     print(h.make_3d())
    <Helioprojective Coordinate (obstime=2020-04-08T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        [(   0., 0., 0.99660825), ( 319., 0., 0.99660944),
            ( 638., 0., 0.99661302), ( 957., 0., 0.99661898),
            (1276., 0., 0.99662732), (1595., 0., 0.99663805),
            (1914., 0., 0.99665116)]>
    """
    screen_type = 'planar'

    @u.quantity_input
    def __init__(self, vantage_point, *, distance_from_center: u.m=0*u.m, **kwargs):
        self._vantage_point = vantage_point
        self._distance_from_center = distance_from_center
        super().__init__(**kwargs)

    def calculate_distance(self, frame):
        obs_to_vantage = self._vantage_point.transform_to(frame).cartesian
        vantage_to_sun = CartesianRepresentation(1, 0, 0) * frame.observer.radius - obs_to_vantage
        direction = vantage_to_sun / vantage_to_sun.norm()
        d_from_plane = frame.observer.radius * direction.x - self._distance_from_center
        rep = frame.represent_as(UnitSphericalRepresentation)
        distance = d_from_plane / rep.dot(direction)

        # Iterate the calculation if differential rotation is being applied
        if _transformations._autoapply_diffrot is not None and np.any(frame.obstime != self._vantage_point.obstime):
            screen_frame = Helioprojective(observer=self._vantage_point, obstime=self._vantage_point.obstime)
            distance = self._iterate_calculate_distance(frame, distance, screen_frame)

        return distance



"""
================================
Generating a map from data array
================================

A simple demonstration of creating a map from a numpy array of data.
"""
import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates import frames

##############################################################################
# Let's create some data.

data = np.arange(0, 100).reshape(10, 10)

##############################################################################
# Next we need to create the metadata. This is made easier using the
# :func:`~sunpy.map.header_helper.make_fitswcs_header` function which will
# create a header object for you. First define the reference coordinate
# which requires a time and an observer location.

coord = SkyCoord(0*u.arcsec, 0*u.arcsec, obstime='2013-10-28 08:24',
                 observer='earth', frame=frames.Helioprojective)

##############################################################################
# Let's pass that into the helper function along with some parameters.
# The reference pixel is the pixel is the one at the reference coordinate. The
# scale sets the size of the pixels. You can also to set a number of other
# metadata as well such as the instrument name and wavelength.

header = sunpy.map.make_fitswcs_header(data, coord,
                                       reference_pixel=[0, 0]*u.pixel,
                                       scale=[2, 2]*u.arcsec/u.pixel,
                                       telescope='Fake Telescope', instrument='UV detector',
                                       wavelength=1000*u.angstrom)

##############################################################################
# Let's now create our map.

manual_map = sunpy.map.Map(data, header)

##############################################################################
# Let's plot the result.

fig = plt.figure()
ax = fig.add_subplot(projection=manual_map)
manual_map.plot(axes=ax)
plt.show()



"""
Ephemeris calculations using SunPy coordinate frames
"""
import re

import numpy as np
import requests

import astropy.units as u
from astropy.constants import c as speed_of_light
from astropy.coordinates import (
    ICRS,
    HeliocentricEclipticIAU76,
    SkyCoord,
    get_body_barycentric,
    get_body_barycentric_posvel,
)
from astropy.coordinates.representation import (
    CartesianDifferential,
    CartesianRepresentation,
    SphericalRepresentation,
)
from astropy.io import ascii
from astropy.time import Time

from sunpy import log
from sunpy.time import parse_time
from sunpy.time.time import _variables_for_parse_time_docstring
from sunpy.util.decorators import add_common_docstring
from .frames import HeliographicStonyhurst

__all__ = ['get_body_heliographic_stonyhurst', 'get_earth',
           'get_horizons_coord']


@add_common_docstring(**_variables_for_parse_time_docstring())
def get_body_heliographic_stonyhurst(body, time='now', observer=None, *, include_velocity=False,
                                     quiet=False):
    """
    Return a `~sunpy.coordinates.frames.HeliographicStonyhurst` frame for the location of a
    solar-system body at a specified time. The location can be corrected for light travel time
    to an observer.

    Parameters
    ----------
    body : `str`
        The solar-system body for which to calculate positions
    time : {parse_time_types}
        Time to use in a parse_time-compatible format
    observer : `~astropy.coordinates.SkyCoord`
        If None, the returned coordinate is the instantaneous or "true" location.
        If not None, the returned coordinate is the astrometric location (i.e., accounts for light
        travel time to the specified observer)
    include_velocity : `bool`, optional
        If True, include the body's velocity in the output coordinate. Defaults to False.
    quiet : `bool`, optional
        If True, the function will not emit logger output for light-travel-time corrections.
        Defaults to False.

    Returns
    -------
    out : `~sunpy.coordinates.frames.HeliographicStonyhurst`
        Location of the solar-system body in the `~sunpy.coordinates.HeliographicStonyhurst` frame

    Notes
    -----
    There is no correction for aberration due to observer motion. For a body close to the Sun in
    angular direction relative to the observer, the correction can be negligible because the
    apparent location of the body will shift in tandem with the Sun.

    For planets other than Earth, Astropy's built-in ephemeris is not as accurate as JPL
    ephemerides, so one can use `astropy.coordinates.solar_system_ephemeris` to switch
    to a JPL ephemeris. See :ref:`astropy-coordinates-solarsystem` for more information, and see
    :ref:`sphx_glr_generated_gallery_units_and_coordinates_venus_transit.py` for an example.

    Examples
    --------
    >>> from sunpy.coordinates.ephemeris import get_body_heliographic_stonyhurst

    Obtain the location of Venus

    >>> get_body_heliographic_stonyhurst('venus', '2012-06-06 04:07:29')
    <HeliographicStonyhurst Coordinate (obstime=2012-06-06T04:07:29.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (0.07349535, 0.05223575, 0.72605496)>

    Obtain the location of Venus as seen from Earth when adjusted for light travel time

    >>> earth = get_body_heliographic_stonyhurst('earth', '2012-06-06 04:07:29')
    >>> get_body_heliographic_stonyhurst('venus', '2012-06-06 04:07:29', observer=earth)
    INFO: Apparent body location accounts for 144.07 seconds of light travel time [sunpy.coordinates.ephemeris]
    <HeliographicStonyhurst Coordinate (obstime=2012-06-06T04:07:29.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (0.07084926, 0.0520573, 0.72605477)>

    Obtain the location and velocity of Mars

    >>> mars = get_body_heliographic_stonyhurst('mars', '2001-02-03', include_velocity=True)
    >>> mars
    <HeliographicStonyhurst Coordinate (obstime=2001-02-03T00:00:00.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (63.03105777, -5.20656151, 1.6251161)
     (d_lon, d_lat, d_radius) in (arcsec / s, arcsec / s, km / s)
        (-0.02323686, 0.00073376, -1.4798387)>
    """
    obstime = parse_time(time)

    if observer is None:
        # If there is no observer, there is not adjustment for light travel time
        emitted_time = obstime
    else:
        observer_icrs = SkyCoord(observer).icrs.cartesian

        # This implementation is modeled after Astropy's `_get_apparent_body_position`
        light_travel_time = 0.*u.s
        emitted_time = obstime
        delta_light_travel_time = 1.*u.s  # placeholder value
        while np.any(np.fabs(delta_light_travel_time) > 1.0e-8*u.s):
            body_icrs = get_body_barycentric(body, emitted_time)
            distance = (body_icrs - observer_icrs).norm()
            delta_light_travel_time = light_travel_time - distance / speed_of_light
            light_travel_time = distance / speed_of_light
            emitted_time = obstime - light_travel_time

        if not quiet:
            if light_travel_time.isscalar:
                ltt_string = f"{light_travel_time.to_value('s'):.2f}"
            else:
                ltt_string = f"{light_travel_time.to_value('s')}"
            log.info(f"Apparent body location accounts for {ltt_string} "
                     "seconds of light travel time")

    if include_velocity:
        pos, vel = get_body_barycentric_posvel(body, emitted_time)
        body_icrs = pos.with_differentials(vel.represent_as(CartesianDifferential))
    else:
        body_icrs = get_body_barycentric(body, emitted_time)

    body_hgs = ICRS(body_icrs).transform_to(HeliographicStonyhurst(obstime=obstime))

    return body_hgs


@add_common_docstring(**_variables_for_parse_time_docstring())
def get_earth(time='now', *, include_velocity=False):
    """
    Return a `~astropy.coordinates.SkyCoord` for the location of the Earth at a specified time in
    the `~sunpy.coordinates.frames.HeliographicStonyhurst` frame. The longitude will be zero by
    definition.

    Parameters
    ----------
    time : {parse_time_types}
        Time to use in a parse_time-compatible format
    include_velocity : `bool`, optional
        If True, include the Earth's velocity in the output coordinate. Defaults to False.

    Returns
    -------
    out : `~astropy.coordinates.SkyCoord`
        Location of the Earth in the `~sunpy.coordinates.frames.HeliographicStonyhurst` frame

    Notes
    -----
    The Earth's velocity in the output coordinate will invariably be negligible in the longitude
    direction because the `~sunpy.coordinates.frames.HeliographicStonyhurst` frame rotates in time
    such that the plane of zero longitude (the XZ-plane) tracks Earth.

    Examples
    --------
    >>> from sunpy.coordinates.ephemeris import get_earth
    >>> get_earth('2001-02-03 04:05:06')
    <SkyCoord (HeliographicStonyhurst: obstime=2001-02-03T04:05:06.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (0., -6.18656962, 0.98567647)>
    >>> get_earth('2001-02-03 04:05:06', include_velocity=True)
    <SkyCoord (HeliographicStonyhurst: obstime=2001-02-03T04:05:06.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (0., -6.18656962, 0.98567647)
     (d_lon, d_lat, d_radius) in (arcsec / s, arcsec / s, km / s)
        (6.42643739e-11, -0.00279484, 0.24968506)>
    >>> get_earth('2001-02-03 04:05:06', include_velocity=True).transform_to('heliocentricinertial')
    <SkyCoord (HeliocentricInertial: obstime=2001-02-03T04:05:06.000): (lon, lat, distance) in (deg, deg, AU)
        (58.41594489, -6.18656962, 0.98567647)
     (d_lon, d_lat, d_distance) in (arcsec / s, arcsec / s, km / s)
        (0.0424104, -0.00279484, 0.2496851)>
    """
    earth = get_body_heliographic_stonyhurst('earth', time=time, include_velocity=include_velocity)

    # Explicitly set the longitude to 0
    earth_repr = SphericalRepresentation(0*u.deg, earth.lat, earth.radius)

    # Modify the representation in the frame while preserving all differentials (e.g., velocity)
    earth = earth.realize_frame(earth_repr.with_differentials(earth.spherical.differentials))

    return SkyCoord(earth)


@add_common_docstring(**_variables_for_parse_time_docstring())
def get_horizons_coord(body, time='now', id_type=None, *, include_velocity=False):
    """
    Queries JPL HORIZONS and returns a `~astropy.coordinates.SkyCoord` for the location of a
    solar-system body at a specified time. This location is the instantaneous or "true" location,
    and is not corrected for light travel time or observer motion.

    Parameters
    ----------
    body : `str`
        The solar-system body for which to calculate positions. One can also use the search form
        linked below to find valid names or ID numbers.
    id_type : `None`, `str`
        Defaults to `None`, which searches major bodies first, and then searches
        small bodies (comets and asteroids) if no major body is found. If
        ``'smallbody'``, the search is limited to only small bodies. If
        ``'designation'``, the search is limited to only small-body designations.
        If ``'name'``, the search is limited to only small-body names. If
        ``'asteroid_name'`` or ``'comet_name'``, the search is limited to only
        asteroid names or only comet names, respectively.
    time : {parse_time_types}, `dict`
        Time to use in a parse_time-compatible format.

        Alternatively, this can be a dictionary defining a range of times and
        dates; the range dictionary has to be of the form
        {{'start': start_time, 'stop': stop_time, 'step':'n[y|d|m]'}}.
        ``start_time`` and ``stop_time`` must be in a parse_time-compatible format,
        and are interpreted as UTC time. ``step`` must be a string with either a
        number and interval length (e.g. for every 10 minutes, ``'10m'``), or a
        plain number for a number of evenly spaced intervals.

    include_velocity : `bool`, optional
        If True, include the body's velocity in the output coordinate. Defaults to False.

    Returns
    -------
    `~astropy.coordinates.SkyCoord`
        Location of the solar-system body

    Notes
    -----
    Be aware that there can be discrepancies between the coordinates returned by JPL HORIZONS,
    the coordinates reported in mission data files, and the coordinates returned by
    `~sunpy.coordinates.get_body_heliographic_stonyhurst`.

    References
    ----------
    * `JPL HORIZONS <https://ssd.jpl.nasa.gov/?horizons>`__
    * `JPL HORIZONS form to search bodies <https://ssd.jpl.nasa.gov/horizons.cgi?s_target=1#top>`__

    Examples
    --------
    .. minigallery:: sunpy.coordinates.get_horizons_coord

    >>> from sunpy.coordinates import get_horizons_coord

    Query the location of Venus

    >>> get_horizons_coord('Venus barycenter', '2001-02-03 04:05:06')  # doctest: +REMOTE_DATA
    INFO: Obtained JPL HORIZONS location for Venus Barycenter (2) [sunpy.coordinates.ephemeris]
    <SkyCoord (HeliographicStonyhurst: obstime=2001-02-03T04:05:06.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (-33.93155836, -1.64998443, 0.71915147)>

    Query the location of the SDO spacecraft

    >>> get_horizons_coord('SDO', '2011-11-11 11:11:11')  # doctest: +REMOTE_DATA
    INFO: Obtained JPL HORIZONS location for Solar Dynamics Observatory (spacecraft) (-136395) [sunpy.coordinates.ephemeris]
    <SkyCoord (HeliographicStonyhurst: obstime=2011-11-11T11:11:11.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (0.01019118, 3.29640728, 0.99011042)>

    Query the location of the SOHO spacecraft via its ID number (-21)

    >>> get_horizons_coord(-21, '2004-05-06 11:22:33')  # doctest: +REMOTE_DATA
    INFO: Obtained JPL HORIZONS location for SOHO (spacecraft) (-21) [sunpy.coordinates.ephemeris]
    <SkyCoord (HeliographicStonyhurst: obstime=2004-05-06T11:22:33.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (0.25234902, -3.55863633, 0.99923086)>

    Query the location and velocity of the asteroid Juno

    >>> get_horizons_coord('Juno', '1995-07-18 07:17', 'smallbody', include_velocity=True)  # doctest: +REMOTE_DATA
    INFO: Obtained JPL HORIZONS location for 3 Juno (A804 RA) [sunpy.coordinates.ephemeris]
    <SkyCoord (HeliographicStonyhurst: obstime=1995-07-18T07:17:00.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, AU)
        (-25.16107532, 14.59098438, 3.17667664)
     (d_lon, d_lat, d_radius) in (arcsec / s, arcsec / s, km / s)
        (-0.03306548, 0.00052415, -2.66709222)>

    Query the location of Solar Orbiter at a set of 12 regularly sampled times

    >>> get_horizons_coord('Solar Orbiter',
    ...                    time={{'start': '2020-12-01',
    ...                           'stop': '2020-12-02',
    ...                           'step': '12'}})  # doctest: +REMOTE_DATA
    INFO: Obtained JPL HORIZONS location for Solar Orbiter (spacecraft) (-144) [sunpy.coordinates.ephemeris]
    ...
    """
    # Reference plane defaults to the ecliptic (IAU 1976 definition)
    args = {
        'EPHEM_TYPE': 'VECTORS',
        'OUT_UNITS': 'AU-D',  # units of AU and days
        'CENTER': '500@10',  # origin is body center (500) of the Sun (10)
        'VEC_TABLE': '2',  # output the 6-element state vector
        'CSV_FORMAT': 'YES',
    }

    if id_type in [None, 'smallbody', 'designation', 'name', 'asteroid_name', 'comet_name']:
        prepend = {
            None: "",
            'smallbody': "",
            'designation': "DES=",
            'name': "NAME=",
            'asteroid_name': "ASTNAM=",
            'comet_name': "COMNAM=",
        }
        if id_type == 'smallbody' and body[-1] != ';':
            body += ';'
        args['COMMAND'] = f"'{prepend[id_type]}{body}'"
    else:
        raise ValueError("Invalid id_type")

    if isinstance(time, dict):
        if set(time.keys()) != set(['start', 'stop', 'step']):
            raise ValueError('time dictionary must have the keys ["start", "stop", "step"]')
        jpl_fmt = "'%Y-%m-%d %H:%M:%S.%f'"
        args['START_TIME'] = parse_time(time['start']).tdb.strftime(jpl_fmt)
        args['STOP_TIME'] = parse_time(time['stop']).tdb.strftime(jpl_fmt)
        args['STEP_SIZE'] = time['step']
    else:
        obstime = parse_time(time)
        if obstime.size >= 10000:
            raise ValueError("For more than 10,000 time values, use dictionary input.")
        array_time = np.reshape(obstime, (-1,))  # Convert to an array, even if scalar
        epochs = array_time.tdb.jd.tolist()  # Time must be provided in JD TDB
        args['TLIST'] = '\n'.join(str(epoch) for epoch in epochs)
        args['TLIST_TYPE'] = 'JD'  # needs to be explicitly set to avoid potential MJD confusion

    contents = "!$$SOF\n" + '\n'.join(f"{k}={v}" for k, v in args.items())
    log.debug(f"JPL HORIZONS query via POST request:\n{contents}")

    output = requests.post('https://ssd.jpl.nasa.gov/api/horizons_file.api',
                           data={'format': 'text'}, files={'input': contents})

    lines = output.text.splitlines()
    start_index, stop_index = 0, len(lines)
    success = False
    error_message = ''
    for index, line in enumerate(lines):
        if line.startswith("Target body name:"):
            target_name = re.search(r': (.*) {', line).group(1)
            log.info(f"Obtained JPL HORIZONS location for {target_name}")

        if "Multiple major-bodies match string" in line:
            error_message = '\n'.join(lines[index:-1])
            break

        if "Matching small-bodies:" in line:
            # Prepare the error message assuming there are multiple matches
            error_message = '\n'.join(lines[index:-1])

            # Change the error message if there are actually zero matches
            if "No matches found." in error_message:
                error_message = "No matches found."

            break

        if line.startswith("$$SOE"):
            start_index = index + 1
        if line.startswith("$$EOE"):
            stop_index = index
            success = True

    if not success:
        if error_message:
            raise ValueError(error_message)
        else:
            raise RuntimeError(f"Unknown JPL HORIZONS error:\n{output.text}")

    column_names = [name.strip() for name in lines[start_index - 3].split(',')]
    result = ascii.read(lines[start_index:stop_index], names=column_names)

    if isinstance(time, dict):
        obstime_tdb = parse_time(result['JDTDB'], format='jd', scale='tdb')
        obstime = Time(obstime_tdb, format='isot', scale='utc')
    else:
        # JPL HORIZONS results are sorted by observation time, so this sorting needs to be undone.
        # Calling argsort() on an array returns the sequence of indices of the unsorted list to put the
        # list in order. Calling argsort() again on the output of argsort() reverses the mapping:
        # the output is the sequence of indices of the sorted list to put that list back in the
        # original unsorted order.
        unsorted_indices = obstime.argsort().argsort()
        result = result[unsorted_indices]

    vector = CartesianRepresentation(result['X'], result['Y'], result['Z']) * u.AU
    if include_velocity:
        velocity = CartesianDifferential(result['VX'], result['VY'], result['VZ']) * u.AU / u.day
        vector = vector.with_differentials(velocity)
    coord = SkyCoord(vector, frame=HeliocentricEclipticIAU76, obstime=obstime)

    return coord.transform_to(HeliographicStonyhurst).reshape(obstime.shape)



def print_missing_dependencies_report(missing, package="sunpy"):
    printed = False
    required_missing = missing.pop("required")
    if required_missing:
        printed = True
        print(f"The following packages are not installed but are required by {package}:")
        for dep in required_missing:
            print(f"* {dep}")
    for extra_name, dependencies in missing.items():
        if "docs" in extra_name:
            continue
        if not dependencies:
            continue
        printed = True
        print(
            f"The following packages are not installed for the {package}[{extra_name}] requirement:")
        for dep in dependencies:
            print(f"  * {dep}")
    return printed


def _self_test_args(*, package=None, online=False, online_only=False, figure_only=False, verbose=False):
    import importlib

    test_args = ["-W", "ignore"]
    if online:
        test_args.append("--remote-data=any")
    if online_only:
        test_args.append("--remote-data=any -m remote_data")
    if package:
        try:
            importlib.import_module(f"sunpy.{package}")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(f"sunpy.{package} was not found.")
        test_args.extend(["--pyargs", f"sunpy.{package}"])
    else:
        test_args.extend(["--pyargs", "sunpy"])
    if figure_only:
        test_args.extend(["-m", "mpl_image_compare"])
    if verbose:
        test_args.append("-vvv")
    return test_args


def self_test(*, package=None, online=False, online_only=False, figure_only=False, verbose=False):
    from sunpy.util.sysinfo import missing_dependencies_by_extra

    print("Starting sunpy self test...")
    print()
    print("Checking for packages needed to run sunpy:")
    missing_deps = missing_dependencies_by_extra(exclude_extras=["asdf", "dask", "dev", "all", "docs"])
    missing_tests = {**missing_deps.pop("core"), **missing_deps.pop("tests-only")}
    printed = print_missing_dependencies_report(missing_deps)
    if not printed:
        print("All required and optional sunpy dependencies are installed.")
    if missing_tests:
        print("You do not have all the required dependencies installed to run the sunpy test suite.")
        print()
        for dep in missing_tests:
            print(f"  * {dep}")
        print()
        print("If are using conda, you will want to run conda install <package name>")
        print('Otherwise you will want run pip install "sunpy[tests]"')
        print()
        return 1
    print()
    print("Starting the sunpy test suite:")
    test_args = _self_test_args(
        package=package, online=online, online_only=online_only, figure_only=figure_only, verbose=verbose
    )

    import pytest

    if verbose:
        print(f"Running pytest with arguments: {test_args}")
    return pytest.main(test_args)


if __name__ == "__main__":
    import sys
    import argparse

    parser = argparse.ArgumentParser(description="Run the sunpy test suite")
    parser.add_argument("--package", help="The sunpy subpackage to test.")
    parser.add_argument("--online", action="store_true",
                        help="Run the tests that require internet access.")
    parser.add_argument("--online-only", action="store_true",
                        help="Run only the tests that require internet access.")
    parser.add_argument("--figure-only", action="store_true",
                        help="Run only the tests that generate figures.")
    parser.add_argument('--verbose', action='store_true', help="Run in verbose mode")
    test_args = parser.parse_args()
    sys.exit(self_test(package=test_args.package, online=test_args.online,
                       online_only=test_args.online_only, figure_only=test_args.figure_only, verbose=test_args.verbose))



"""
===================================================
Obtaining a spacecraft trajectory from JPL Horizons
===================================================

This example shows how to obtain the trajectory of a spacecraft from JPL Horizons
and plot it relative to other bodies in the solar system.

JPL `Horizons <https://ssd.jpl.nasa.gov/horizons/>`__ can return the locations of
planets and minor bodies (e.g., asteroids) in the solar system, and it can also
return the location of a variety of major spacecraft.
"""

import matplotlib.pyplot as plt

import astropy.units as u

from sunpy.coordinates import get_body_heliographic_stonyhurst, get_horizons_coord
from sunpy.time import parse_time

##############################################################################
# We use :func:`~sunpy.coordinates.get_horizons_coord` to query JPL Horizons
# for the trajectory of Parker Solar Probe (PSP). Let's request 50 days on
# either side of PSP's 14th closest approach to the Sun.

perihelion_14 = parse_time('2022-12-11 13:16')
psp = get_horizons_coord('Parker Solar Probe',
                         {'start': perihelion_14 - 50 * u.day,
                          'stop': perihelion_14 + 50 * u.day,
                          'step': '180m'})

##############################################################################
# We also obtain the location of Earth at PSP perihelion. We could query
# JPL Horizons again, but :func:`~sunpy.coordinates.get_body_heliographic_stonyhurst` returns
# a comparably accurate location using the Astropy ephemeris.

earth = get_body_heliographic_stonyhurst('Earth', perihelion_14)

##############################################################################
# For the purposes of plotting on a Matplotlib polar plot, we create a short
# convenience function to extract the necessary values in the appropriate units.


def coord_to_polar(coord):
    return coord.lon.to_value('rad'), coord.radius.to_value('AU')

##############################################################################
# Finally, we plot the trajectory on a polar plot. Be aware that the
# orientation of the Stonyhurst heliographic coordinate system rotates
# over time such that the Earth is always at zero longitude.
# Accordingly, when we directly plot the trajectory, it does not appear
# as a simple ellipse because each trajectory point has a different
# observation time and thus a different orientation of the coordinate
# system. To see the elliptical orbit, the trajectory can be
# transformed to the coordinate frame of Earth at the single time of
# PSP perihelion (``earth``), so that the trajectory is represented in
# a non-rotating coordinate frame.


fig = plt.figure()
ax = fig.add_subplot(projection='polar')
ax.plot(0, 0, 'o', label='Sun', color='orange')
ax.plot(*coord_to_polar(earth), 'o', label='Earth', color='blue')
ax.plot(*coord_to_polar(psp),
        label='PSP (as seen from Earth)', color='purple')
ax.plot(*coord_to_polar(psp.transform_to(earth)),
        label='PSP (non-rotating frame)', color='purple', linestyle='dashed')
ax.set_title('Stonyhurst heliographic coordinates')
ax.legend(loc='upper center')

plt.show()

##############################################################################
# There are other tools that enable a similar style of figure.
# `solarmach <https://github.com/jgieseler/solarmach#usage>`__ is one such example.



"""
Test cases for AIAMap subclass.
"""
import copy

import pytest

import astropy.units as u
from astropy.visualization.mpl_normalize import ImageNormalize

from sunpy.data.test import get_test_filepath
from sunpy.map import Map
from sunpy.map.sources.sdo import AIAMap
from sunpy.tests.helpers import SKIP_GLYMUR
from .helpers import _test_private_date_setters

params = [get_test_filepath("aia_171_level1.fits")]
if not SKIP_GLYMUR:
    params += [get_test_filepath("2013_06_24__17_31_30_84__SDO_AIA_AIA_193.jp2")]


@pytest.fixture(scope="module", params=params)
def aia_map(request):
    return Map(request.param)


def test_aia_map(aia_map):
    """Tests the creation of AIAMap"""
    assert isinstance(aia_map, AIAMap)


def test_reference_date(aia_map):
    assert aia_map.reference_date.isot in ["2011-02-15T00:00:01.340", "2013-06-24T17:31:31.840"]


def test_date(aia_map):
    assert aia_map.date.isot in ["2011-02-15T00:00:00.340", "2013-06-24T17:31:30.840"]


def test_private_date_setters(aia_map):
    _test_private_date_setters(aia_map)


def test_is_datasource_for(aia_map):
    """Tests the is_datasource_for method of AIAMap."""
    assert aia_map.is_datasource_for(aia_map.data, aia_map.meta)


def test_observatory(aia_map):
    """Tests the observatory property of the AIAMap object."""
    assert aia_map.observatory == "SDO"


def test_measurement(aia_map):
    """Tests the measurement property of the AIAMap object."""
    # aiaimg has 171, jp2path has 193.
    assert aia_map.measurement.value in [171, 193]


def test_norm_clip(aia_map):
    # Tests that the default normalizer has clipping disabled
    assert not aia_map.plot_settings['norm'].clip


def test_new_instance_preserves_plot_settings(aia_map):
    # Tests that the _new_instance method preserves the plot_settings
    # of the old instance. This is done on the AIA source as the AIAMap
    # constructor explicitly sets the cmap and norm and we want to test
    # that _new_instance persists the old custom plot_settings
    aia_map.plot_settings['norm'] = ImageNormalize(vmin=0.1, vmax=42)
    aia_map.plot_settings['cmap'] = 'inferno'
    new_aia_map = aia_map._new_instance(aia_map.data,
                                        aia_map.meta,
                                        plot_settings=aia_map.plot_settings)
    assert new_aia_map.plot_settings['norm'].vmin == aia_map.plot_settings['norm'].vmin
    assert new_aia_map.plot_settings['norm'].vmax == aia_map.plot_settings['norm'].vmax
    assert new_aia_map.plot_settings['cmap'] == aia_map.plot_settings['cmap']
    # If no plot settings are explicitly passed, the plot_settings should fall back to those
    # in the constructor
    new_aia_map = aia_map._new_instance(aia_map.data, aia_map.meta)
    assert new_aia_map.plot_settings['norm'].vmin is None
    assert new_aia_map.plot_settings['norm'].vmax is None
    assert new_aia_map.plot_settings['cmap'] == new_aia_map._get_cmap_name()


def test_wcs(aia_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    aia_map.pixel_to_world(0*u.pix, 0*u.pix)


def test_missing_tobs(aia_map):
    # Should fall back to base reference_date for reference date if T_OBS is missing
    new_meta = copy.deepcopy(aia_map.meta)
    new_meta.pop('T_OBS')
    new_aia_map = aia_map._new_instance(aia_map.data, new_meta)
    assert new_aia_map.reference_date == super(type(new_aia_map), new_aia_map).reference_date



"""
===========================================
Overlaying off-disk contours using a screen
===========================================

This example shows how to use a screen to overlay off-disk contours from one
map onto another map.

When overlaying contours from one data set onto another data set, there is
almost always some small difference in the coordinate frame of the two data
sets, namely in the observation time and the observer location. With the
slight shift in perspective, any off-disk contours are not plotted by default
due to the lack of knowledge of where the contour exists in 3D space. (In
contrast, on-disk contours do not have this issue because the default
assumption is that on-disk contours lie on the surface of the Sun.) To change
the behavior for off-disk contours, we can use a screen to specify an
assumption for the 3D location of 2D off-disk points.

See :ref:`sphx_glr_generated_gallery_map_transformations_reprojection_spherical_screen.py`
for a different use case for screens.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
import sunpy.map
from sunpy.coordinates.screens import SphericalScreen

##############################################################################
# We use two AIA images that were not taken at exactly the same time. We
# crop the maps to an active region at the eastern limb.

aia171_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
aia193_map = sunpy.map.Map(sunpy.data.sample.AIA_193_IMAGE)
aia171_map = aia171_map.submap((0, 500)*u.pix, top_right=(250, 750)*u.pix)
aia193_map = aia193_map.submap((0, 500)*u.pix, top_right=(250, 750)*u.pix)

##############################################################################
# First, let's plot the AIA 171 contours on their native map. The contours
# extend off disk due to the coronal loops.

fig = plt.figure()
ax = fig.add_subplot(projection=aia171_map)
aia171_map.plot(axes=ax)
aia171_map.draw_contours(axes=ax, levels=[1000, 2500] * u.DN, colors="green")

##############################################################################
# Next, let's try overlaying those same AIA 171 contours on the AIA 193 map.
# Note that none of the off-disk contours appear. Again, this is because it
# is not well defined where to plot such contours from the vantage point of
# the AIA 193 map.

fig = plt.figure()
ax = fig.add_subplot(projection=aia193_map)
aia193_map.plot(axes=ax, title="Default overlay behavior")
aia171_map.draw_contours(axes=ax, levels=[1000, 2500] * u.DN, colors="green")

##############################################################################
# Finally, let's overlay those contours again, but this time using the
# `~sunpy.coordinates.SphericalScreen` context manager. This screen assumes
# that off-disk points lie on the inside of a spherical screen. Note that the
# off-disk contours are now visible.

fig = plt.figure()
ax = fig.add_subplot(projection=aia193_map)
aia193_map.plot(axes=ax, title="Using SphericalScreen")
with SphericalScreen(aia171_map.observer_coordinate, only_off_disk=True):
    aia171_map.draw_contours(axes=ax, levels=[1000, 2500] * u.DN, colors="green")

plt.show()

# sphinx_gallery_thumbnail_number = 3



"""
This module implements a SRS File Reader.
"""
import re
import datetime
from collections import OrderedDict

import numpy as np

import astropy.io.ascii
import astropy.units as u
from astropy.table import Column, MaskedColumn, QTable, vstack

__all__ = ['read_srs']


def read_srs(filepath):
    """
    Parse a SRS table from NOAA SWPC.

    Parameters
    ----------
    filepath : `str`
        The full path to a SRS table.

    Returns
    -------
    table : `astropy.table.QTable`
        Table containing a stacked table from all the tables in the SRS file.
        The header information is stored in the ``.meta`` attribute.
    """
    with open(filepath) as srs:
        file_lines = srs.readlines()

    header, section_lines, supplementary_lines = split_lines(file_lines)

    return make_table(header, section_lines, supplementary_lines)


def make_table(header, section_lines, supplementary_lines):
    """
    From the separated section lines and the header, clean up the data and
    convert to a `~astropy.table.QTable`.
    """
    meta_data = get_meta_data(header, supplementary_lines)

    tables = []
    for i, lines in enumerate(section_lines):
        if lines:
            key = list(meta_data['id'].keys())[i]
            t1 = astropy.io.ascii.read(lines)

            # Change column names into titlecase
            column_names = list(t1.columns)
            t1.rename_columns(column_names, new_names=[col.title() for col in column_names])

            if len(t1) == 0:
                col_data_types = {
                    # ID : <class 'str'>
                    'Nmbr': np.dtype('i4'),
                    'Location': np.dtype('U6'),
                    'Lo': np.dtype('i8'),
                    'Area': np.dtype('i8'),
                    'Z': np.dtype('U3'),
                    'Ll': np.dtype('i8'),
                    'Nn': np.dtype('i8'),
                    'Magtype': np.dtype('S4'),
                    'Lat': np.dtype('i8'),
                }
                for c in t1.itercols():
                    # Put data types of columns in empty table to correct types,
                    # or else vstack will fail.
                    c.dtype = col_data_types[c._name]
                t1.add_column(
                    Column(data=None, name="ID", dtype=('S2')), index=0)
            else:
                t1.add_column(Column(data=[key] * len(t1), name="ID"), index=0)

            tables.append(t1)

    out_table = vstack(tables)

    # Parse the Location column in Table 1
    if 'Location' in out_table.columns:
        col_lat, col_lon = parse_location(out_table['Location'])
        del out_table['Location']
        out_table.add_column(col_lat)
        out_table.add_column(col_lon)

    # Parse the Lat column in Table 3
    if 'Lat' in out_table.columns:
        parse_lat_col(out_table['Lat'], out_table['Latitude'])
        del out_table['Lat']

    # Give columns more sensible names
    column_mapping = {
        'Nmbr': 'Number',
        'Nn': 'Number of Sunspots',
        'Lo': 'Carrington Longitude',
        'Magtype': 'Mag Type',
        'Ll': 'Longitudinal Extent',
    }

    for old_name, new_name in column_mapping.items():
        out_table.rename_column(old_name, new_name)

    # Define a Solar Hemisphere Unit
    a = {}
    u.def_unit(
        "SH",
        represents=(2 * np.pi * u.solRad**2),
        prefixes=True,
        namespace=a,
        doc="A solar hemisphere is the area of the visible solar disk.")

    # Set units on the table
    out_table['Carrington Longitude'].unit = u.deg
    out_table['Area'].unit = a['uSH']
    out_table['Longitudinal Extent'].unit = u.deg

    out_table.meta = meta_data

    # Number should be formatted in 10000 after 2002-06-15.
    if out_table.meta['issued'] > datetime.datetime(2002, 6, 15):
        out_table['Number'] += 10000

    return QTable(out_table)


def split_lines(file_lines):
    """
    Given all the lines in the file split based on the three sections and
    return the lines for the header, a list of lines for each section that
    is not 'None', and a list of supplementary lines after the main sections
    if not 'None'.
    """
    section_lines = []
    final_section_lines = []
    for i, line in enumerate(file_lines):
        if re.match(r'^(I\.|IA\.|II\.)', line):
            section_lines.append(i)
        if re.match(r'^(III|COMMENT|EFFECTIVE 2 OCT 2000|PLAIN|This message is for users of the NOAA/SEC Space|NNN)', line, re.IGNORECASE):
            final_section_lines.append(i)

    if final_section_lines and final_section_lines[0] > section_lines[-1]:
        section_lines.append(final_section_lines[0])

    header = file_lines[:section_lines[0]]
    header += [file_lines[s] for s in section_lines]

    # Append comments to the comment lines
    for line in section_lines:
        file_lines[line] = '# ' + file_lines[line]
    t1_lines = file_lines[section_lines[0]:section_lines[1]]
    # Remove the space so table reads it correctly
    t1_lines[1] = re.sub(r'Mag\s*Type', r'Magtype', t1_lines[1], flags=re.IGNORECASE)
    t2_lines = file_lines[section_lines[1]:section_lines[2]]

    # SRS files before 2000-10-02 files may have an empty `COMMENT` column in ``t2_lines``
    if "COMMENT" in t2_lines[1].split():
        expected_pattern_dict = {
            'Nmbr': r'^\d+$',
            'Location': r'^(?:[NESW](?:\d{2})){1,2}$',
            'Lo': r'^\d+$',
        }
        # Try to drop the comment column and return in original format
        t2_lines[1:] = _try_drop_empty_column("COMMENT", t2_lines[1:], expected_pattern_dict)

    if len(section_lines) > 3:
        t3_lines = file_lines[section_lines[2]:section_lines[3]]
        supplementary_lines = file_lines[section_lines[3]:]
    else:
        t3_lines = file_lines[section_lines[2]:]
        supplementary_lines = None

    lines = [t1_lines, t2_lines, t3_lines]
    for i, ll in enumerate(lines):
        if len(ll) > 2 and ll[2].strip().title() == 'None':
            del ll[2]

    return header, lines, supplementary_lines


def get_meta_data(header, supplementary_lines):
    """
    Convert a list of header lines and a list of supplementary lines (if not 'None') into a meta data dict.
    """
    meta_lines = []
    for line in header:
        if line.startswith(':'):
            meta_lines.append(line)

    meta_data = {}
    for m in meta_lines:
        if re.search(r'Corrected\s*Copy', m, re.IGNORECASE):
            meta_data['corrected'] = True
            continue
        k, v = m.strip().split(':')[1:]
        meta_data[k.lower()] = v.strip()
    meta_data['issued'] = datetime.datetime.strptime(meta_data['issued'],
                                                     "%Y %b %d %H%M UTC")

    # Get ID descriptions
    meta_data['id'] = OrderedDict()
    for h in header:
        if h.startswith(("I.", "IA.", "II.")):
            i = h.find('.')
            k = h[:i]
            v = h[i + 2:]
            meta_data['id'][k] = v.strip()

    meta_data['header'] = [h.strip() for h in header]

    if supplementary_lines:
        meta_data['supplementary_lines'] = [sl.strip() for sl in supplementary_lines]

    return meta_data


def parse_longitude(value):
    """
    Parse longitude in the form "W10" or "E10".
    """
    lonsign = {'W': 1, 'E': -1}
    if "W" in value or "E" in value:
        return lonsign[value[3]] * float(value[4:])


def parse_latitude(value):
    """
    Parse latitude in the form "S10" or "N10".
    """
    latsign = {'N': 1, 'S': -1}
    if "N" in value or "S" in value:
        return latsign[value[0]] * float(value[1:3])


def parse_location(column):
    """
    Given a column of location data in the form "S10E10" convert to two columns
    of angles.
    """
    latitude = MaskedColumn(name="Latitude", unit=u.deg)
    longitude = MaskedColumn(name="Longitude", unit=u.deg)

    for i, loc in enumerate(column):
        if loc:
            lati = parse_latitude(loc)
            longi = parse_longitude(loc)
            latitude = latitude.insert(i, lati)
            longitude = longitude.insert(i, longi)
        else:
            latitude = latitude.insert(i, None, mask=True)
            longitude = longitude.insert(i, None, mask=True)
    return latitude, longitude


def parse_lat_col(column, latitude_column):
    """
    Given an input column of "latitudes" in the form "S10" parse them and add
    them to an existing column of "latitudes".
    """
    for i, loc in enumerate(column):
        if loc:
            latitude_column.mask[i] = False
            latitude_column[i] = parse_latitude(loc)
    return latitude_column


def _try_drop_empty_column(column_name_to_drop, data_lines, pattern_dict):
    """
    Try dropping an empty ``column_name_to_drop`` from ``data_lines``.

    Parameters
    ----------
    column_name_to_drop : `str`
        Name of the empty column to be dropped.
    data_lines : `list[str]`
        List of lines extracted from a file (each line is a string)
        corresponding to the header (e.g. ``header = data_lines[0]``)
        and the data (``data = data_lines[1:]``)
    pattern_dict : `dict`
        A dictionary specifying the patterns to match for each column

    Returns
    -------
    `list[str]`
        The modified ``data_lines`` in titlecase with the specified column dropped, if all validations pass.

    """
    # Create a lowercase pattern dict
    pattern_dict_lower = {key.lower(): value for key, value in pattern_dict.items()}

    # Extract columns and rows
    header_line, *row_lines = data_lines
    column_list = [column.strip().lower() for column in header_line.split()]

    # Drop ``column_name_to_drop`` if exists
    try:
        column_index = column_list.index(column_name_to_drop.strip().lower())
        column_list.pop(column_index)
    except ValueError:
        raise ValueError(f"The column '{column_name_to_drop}' does not exist.")

    # Remove the dropped column from pattern_dict
    pattern_dict_lower.pop(column_name_to_drop.strip().lower(), None)

    # If the data is `None`, just return the header/data
    if row_lines[0].strip().title() == 'None':
        # Return as titlecase
        column_list = [col.title() for col in column_list]
        return [" ".join(column_list)] + row_lines

    # Check if the remaining columns are a subset of the columns in pattern_dict
    remaining_columns_set = set(column_list)
    pattern_columns_set = set(pattern_dict_lower.keys())
    if not remaining_columns_set.issubset(pattern_columns_set):
        raise ValueError("The remaining columns are not a subset of the columns in ``pattern_dict``.")

    # Check if all rows have the same length as the remaining columns
    row_lengths_equal = all(len(row.split()) == len(column_list) for row in row_lines)
    if not row_lengths_equal:
        raise ValueError("not all rows have the same number of values as the remaining columns.")

    # Check that the row values are consistent with the provided pattern dictionary
    matching_pattern = all(all(re.match(pattern_dict_lower[column], value) for column, value in zip(column_list, row.split())) for row in row_lines)
    if not matching_pattern:
        raise ValueError("not all rows match the provided pattern.")

    # Return as titlecase
    column_list = [col.title() for col in column_list]
    return [" ".join(column_list)] + row_lines



"""
==========================
Plotting a coordinate grid
==========================

This example demonstrates how you can draw a grid of solar coordinates on top
of a map.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
import sunpy.map

################################################################################
# Let's start with a sample AIA image.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)

################################################################################
# Now we can plot the image, and add grids in heliographic Stonyhurst and
# heliographic Carrington coordinates.

fig = plt.figure()
ax1 = fig.add_subplot(121, projection=aia_map)
aia_map.plot(axes=ax1, clip_interval=(1, 99.9)*u.percent)
stonyhurst_grid = aia_map.draw_grid(axes=ax1, system='stonyhurst')

ax2 = fig.add_subplot(122, projection=aia_map)
aia_map.plot(axes=ax2, clip_interval=(1, 99.9)*u.percent)
carrington_grid = aia_map.draw_grid(axes=ax2, system='carrington')

for ax in [ax1, ax2]:
    ax.set_title('')
    ax.set_xlim(0, 400)
    ax.set_ylim(0, 400)
    # Turn of the map grid and ticks
    ax.grid(False)
    for coord in [0, 1]:
        ax.coords[coord].set_ticklabel_visible(False)
        ax.coords[coord].set_ticks_visible(False)

stonyhurst_grid['lon'].set_ticks([-30, -45, -60] * u.deg)
carrington_grid['lon'].set_ticks([-45, -60, -75] * u.deg)

fig.subplots_adjust(wspace=0.5)

plt.show()



"""
This module provides time formats specific to solar physics
"""
from astropy.time.formats import TimeFromEpoch, erfa

__all__ = ['TimeUTime', 'TimeTaiSeconds']


class TimeUTime(TimeFromEpoch):
    """
    UT seconds from 1979-01-01 00:00:00 UTC, ignoring leap seconds.

    Notes
    -----
    This format "ignores" leap seconds by treating each day as spanned by exactly
    86400 seconds, which means that this format's second is not actually uniform in
    duration. On a day without a leap second, this format's second is equal to an
    SI second. On a day with a leap second, this format's second is larger than an
    SI second by 1/86400 of an SI second.

    This format is very similar to the default output format of the ``anytim``
    routine in SSW in that there are exactly 86400 seconds assigned for each day.
    However, ``anytim`` treats the seconds as always equal to an SI second, and thus
    the 86400 seconds span only the first 86400/86401 of the day, and the leap
    second is skipped over. This results in discrepancies of up to a second on days
    with a leap second.

    This format is equivalent to `~astropy.time.TimeUnix`, except that the epoch is
    9 years later.

    References
    ----------
    * `anytim routine in SSW <https://hesperia.gsfc.nasa.gov/ssw/gen/idl/utplot/anytim.pro>`__

    Examples
    --------
    >>> from astropy.time import Time
    >>> t = Time('2000-01-01T13:53:23')
    >>> print(t.utime)
    662738003.0
    >>> t2 = Time('1979-01-01T00:00:00')
    >>> print(t2.utime)
    0.0
    """
    name = 'utime'
    unit = 1.0 / erfa.DAYSEC  # in days (1 day == 86400 seconds)
    epoch_val = '1979-01-01 00:00:00'
    epoch_val2 = None
    epoch_scale = 'utc'  # Scale for epoch_val class attribute
    epoch_format = 'iso'  # Format for epoch_val class attribute


class TimeTaiSeconds(TimeFromEpoch):
    """
    SI seconds from 1958-01-01 00:00:00, which includes UTC leap seconds.

    1958-01-01 00:00:00 is the defined time when International Atomic Time (TAI)
    and Universal Time (UT) are synchronized. A TAI second has the same length as
    an SI second, but prior to 1972-01-01, a UT second -- then defined to be 1/86400
    of an Earth day -- grew to be longer than than an SI second. 1972-01-01
    00:00:00 UTC is equal to 1972-01-01 00:00:10 TAI. After 1972-01-01, Coordinated
    Universal Time (UTC) is defined with seconds that are the same length as SI
    seconds, but now leap seconds are occasionally added to UTC so that it stays
    synchronized with Earth days.

    Notes
    -----
    This format is equivalent to the output of the SSW ``anytim2tai`` routine, and
    related routines, for times after 1972-01-01. Be aware that the SSW routines
    are not written to provide valid results for times before 1972-01-01.

    This format is equivalent to `~astropy.time.TimeUnixTai`, except that the epoch
    is 12 years earlier.

    References
    ----------
    * `anytim2tai routine in SSW <https://hesperia.gsfc.nasa.gov/ssw/gen/idl/time/anytim2tai.pro>`__

    Examples
    --------
    >>> from astropy.time import Time
    >>> t = Time('1958-01-01T00:00:00', format='isot', scale='tai')
    >>> t.tai_seconds  # doctest: +SKIP
    0.0
    >>> t2 = Time('2015-10-25T05:24:08', format='isot', scale='tai')
    >>> t2.tai_seconds  # doctest: +SKIP
    1824441848.0
    >>> t3 = Time(t2.tai_seconds, format='tai_seconds')  # scale is automatically TAI
    >>> t3.isot
    '2015-10-25T05:24:08.000'
    """
    name = 'tai_seconds'
    unit = 1.0 / erfa.DAYSEC  # in days (1 day == 86400 seconds)
    epoch_val = '1958-01-01 00:00:00'
    epoch_val2 = None
    epoch_scale = 'tai'
    epoch_format = 'iso'



"""
==========================================
Extracting intensity of a map along a line
==========================================

In this example we will extract the intensity values of all the pixels
that intersect with a given set of coordinates.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.data.sample import AIA_171_ROLL_IMAGE

###############################################################################
# First we construct a map, using some sample data.
aia_map = sunpy.map.Map(AIA_171_ROLL_IMAGE)


###############################################################################
# Next we define a path in a `~astropy.coordinates.SkyCoord` object.
# In this example we are just going to use a straight line defined by two points.
# However a path with any number of points and any shape can be used.
line_coords = SkyCoord([-1024, -908], [20, 633], unit=(u.arcsec, u.arcsec),
                       frame=aia_map.coordinate_frame)


###############################################################################
# Next we call the :func:`sunpy.map.pixelate_coord_path` function with the map
# and the coordinate path to obtain the coordinates of the map pixels that
# intersect that path. We pass those coordinates to
# :func:`sunpy.map.sample_at_coords` to extract the values for those map
# pixels.
intensity_coords = sunpy.map.pixelate_coord_path(aia_map, line_coords)
intensity = sunpy.map.sample_at_coords(aia_map, intensity_coords)


###############################################################################
# Next we will calculate the angular separation between the first point and
# every other coordinate we extracted. We are doing this to give us a
# meaningful x-axis for our line plot below.
angular_separation = intensity_coords.separation(intensity_coords[0]).to(u.arcsec)

###############################################################################
# Finally let's plot the results.
fig = plt.figure(figsize=(10, 4))
ax1 = fig.add_subplot(121, projection=aia_map)
aia_map.plot(axes=ax1)
ax1.plot_coord(intensity_coords)
ax1.plot_coord(line_coords[0], marker="o", color="blue", label="start")
ax1.plot_coord(line_coords[1], marker="o", color="green", label="end")
ax1.legend()

ax2 = fig.add_subplot(122)
ax2.plot(angular_separation, intensity)
ax2.set_xlabel("Angular distance along slit [arcsec]")
ax2.set_ylabel(f"Intensity [{aia_map.unit}]")

plt.show()



"""
=============================================
Reproducing the "Where is STEREO Today?" plot
=============================================

How to reproduce the "Where is STEREO Today?" plot.

This example uses capabilities in sunpy to reproduce the plot that can be found
at `<https://stereo-ssc.nascom.nasa.gov/where.shtml>`__. This example is not
written as a tutorial, but there are related tutorials for learning about these
capabilities:

* :ref:`sphx_glr_generated_gallery_units_and_coordinates_planet_locations.py`
* :ref:`sphx_glr_generated_gallery_units_and_coordinates_ParkerSolarProbe_trajectory.py`
"""
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MultipleLocator

import astropy.units as u
from astropy.coordinates import Longitude

from sunpy.coordinates import HeliocentricEarthEcliptic, get_body_heliographic_stonyhurst, get_horizons_coord
from sunpy.time import parse_time

##############################################################################
# Define the time for the plot as the time when this script is run.

# This time is fixed as the ephemeris information for STEREO-B is no longer generated
obstime = parse_time('2024-Oct-17 20:23')

##############################################################################
# Define a convenience function to extract the first full orbit from a
# trajectory, assuming that the trajectory moves in the direction of positive
# ecliptic longitude.

hee_frame = HeliocentricEarthEcliptic(obstime=obstime)


def get_first_orbit(coord):
    lon = coord.transform_to(hee_frame).spherical.lon
    shifted = Longitude(lon - lon[0])
    ends = np.flatnonzero(np.diff(shifted) < 0)
    if ends.size > 0:
        return coord[:ends[0]]
    return coord


##############################################################################
# Obtain the locations and trajectories of the various planets and spacecraft.
# To ensure that each trajectory contains at least one full orbit, we request
# 700 days for each planet and 1 year for each spacecraft.

planets = ['Mercury', 'Venus', 'Earth', 'Mars']
times = obstime + np.arange(700) * u.day
planet_coords = {planet: get_first_orbit(get_body_heliographic_stonyhurst(planet, times))
                 for planet in planets}

stereo_a = get_horizons_coord('STEREO-A', obstime)
stereo_b = get_horizons_coord('STEREO-B', obstime)

missions = ['Parker Solar Probe', 'Solar Orbiter', 'BepiColombo']
mission_labels = {'Parker Solar Probe': 'PSP', 'Solar Orbiter': 'SO', 'BepiColombo': 'BEPICOLOMBO'}
mission_coords = {mission: get_first_orbit(get_horizons_coord(mission, {'start': obstime,
                                                                        'stop': obstime + 1 * u.yr,
                                                                        'step': '1d'}))
                  for mission in missions}

##############################################################################
# Define a convenience function for converting coordinates to plot positions
# in the ecliptic plane.


def coord_to_heexy(coord):
    coord = coord.transform_to(hee_frame)
    coord.representation_type = 'cartesian'
    return coord.y.to_value('AU'), coord.x.to_value('AU')


##############################################################################
# Set Matplotlib settings to the desired appearance and initialize the axes.

mpl.rcParams.update({'figure.facecolor': 'black',
                     'axes.edgecolor': 'white',
                     'axes.facecolor': 'black',
                     'axes.labelcolor': 'white',
                     'axes.titlecolor': 'white',
                     'lines.linewidth': 1,
                     'xtick.color': 'white',
                     'xtick.direction': 'in',
                     'xtick.top': True,
                     'ytick.color': 'white',
                     'ytick.direction': 'in',
                     'ytick.right': True})

fig = plt.figure()
ax = fig.add_subplot()

ax.set_xlim(-2.15, 2.15)
ax.set_xlabel('Y (HEE)')
ax.xaxis.set_major_locator(MultipleLocator(1))
ax.xaxis.set_minor_locator(MultipleLocator(0.1))

ax.set_ylim(1.8, -1.8)
ax.set_ylabel('X (HEE)')
ax.yaxis.set_major_locator(MultipleLocator(1))
ax.yaxis.set_minor_locator(MultipleLocator(0.1))

ax.set_title(obstime.strftime('%d-%b-%Y %H:%M UT'))
ax.set_aspect('equal')

# sphinx_gallery_defer_figures

##############################################################################
# Draw the Sun-Earth line.

ax.plot([0, 0], [0, 2], linestyle='dotted', color='gray')

# sphinx_gallery_defer_figures

##############################################################################
# Draw Mercury, Venus, Earth, and Mars, with Earth formatted differently.

for planet, coord in planet_coords.items():
    ax.plot(*coord_to_heexy(coord), linestyle='dashed', color='gray')

    if planet == 'Earth':
        color, markersize, offset = 'lime', 10, 0.1
    else:
        color, markersize, offset = 'gray', None, 0.05

    x, y = coord_to_heexy(coord[0])
    ax.plot(x, y, 'o', markersize=markersize, color=color)
    ax.text(x + offset, y, planet, color=color)

# sphinx_gallery_defer_figures

##############################################################################
# Draw the STEREO spacecraft (without orbits), as well as Sun-STEREO lines.

for stereo, label, color in [(stereo_a, 'A', 'red'), (stereo_b, 'B', 'blue')]:
    x, y = coord_to_heexy(stereo)
    ax.plot([0, 5*x], [0, 5*y], linestyle='dotted', color='gray')
    ax.plot(x, y, 'o', color=color)
    ax.text(x + 0.1, y, label, color=color, fontsize=18)

# sphinx_gallery_defer_figures

##############################################################################
# Draw the Sun, which is at the origin by definition.

ax.plot(0, 0, 'o', markersize=15, color='yellow')
ax.text(0.12, 0, 'Sun', color='yellow')

# sphinx_gallery_defer_figures

##############################################################################
# Finally, draw the various spacecraft, with Solar Orbiter colored differently.

for mission, coord in mission_coords.items():
    color = 'magenta' if mission == 'Solar Orbiter' else 'orange'

    ax.plot(*coord_to_heexy(coord), linestyle='dashed', color=color)

    x, y = coord_to_heexy(coord[0])
    ax.plot(x, y, 'o', color=color)
    ax.text(x + 0.05, y, mission_labels[mission], color=color)

plt.show()

# This is necessary to reset the Matplotlib settings after plotting for our documentation.
# You don't need this in your own code.
mpl.rcParams.update(mpl.rcParamsDefault)
mpl.rcParams.update({'axes.titlecolor': 'black'})
